/*
 * IMPORTANT: This program must be setuid root and must not be runnable by
 * any user other than the trusted globus service (or root); otherwise
 * regular users could add/delete/modify grid user accounts.  To protect
 * against security breaches due to improperly set file permissions, we
 * refuse to run if the uid is not either (a) 0 or (b) the same as the gid
 * owning the executable.
 * Recommended file permissions/ownership: chmod 4750 && chown root.globus
 */

#include <stdio.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/fsuid.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <grp.h>

#define __USE_GNU                  // O_NOFOLLOW
#include <fcntl.h>
#include <sys/file.h>

// Define how grid account usernames are formed.
#define GRID_USER_PREFIX  "grid"   // must begin with a letter, 1-7 chars long
#define GRID_USER_PREFIX_LEN 4     // strlen(GRID_USER_PREFIX)
#define GRID_UNUM_MIN     0
#define GRID_UNUM_MAX     9999     // prefix+unum_max must be at most 8 chars
#define GRID_UNAME_FORMAT "%s%04d" // how to name new accounts - 8 chars max

// Define the UNIX uid range to use for grid accounts.  This program
// refuses to modify any account outside this range.
#define GRID_UID_MIN      3000     // smallest uid to use
#define GRID_UID_MAX      4999     // largest uid to use

// Primary user group and default shell to assign new accounts.
#define GRID_USER_GROUP   "griduser"
#define GRID_USER_SHELL   "/sbin/nologin"

// Executables used.
#define RM_PROG           "/bin/rm"
#define RMDIR_PROG        "/bin/rmdir"
#define FIND_PROG         "/usr/bin/find"
#define QUOTA_PROG        "/usr/bin/quota"
#define USERADD_PROG      "/usr/sbin/useradd"
#define USERDEL_PROG      "/usr/sbin/userdel"
#define USERMOD_PROG      "/usr/sbin/usermod"
#define SETQUOTA_PROG     "/usr/sbin/setquota"

// This file tracks the last uid/unum assigned, and also serves as an advisory
// lock file for concurrency control -- if multiple instances of this program
// create accounts concurrently, there is otherwise a TOCTTOU condition.
// IMPORTANT: this file should either be pre-created with appropriate
// permissions, or else not in a world-writable directory (else DoS possible).
#define INFO_LOCKFILE     "/tmp/ssinfo.lck"

// Name of the grid-mapfile to modify when creating or removing accounts
#define GRIDMAP           "/etc/grid-security/grid-mapfile"

#define NEXT_GRID_UID(a)  ((a) >= GRID_UID_MAX ? GRID_UID_MIN : (a) + 1)
#define NEXT_GRID_UNUM(a) ((a) >= GRID_UNUM_MAX ? GRID_UNUM_MIN : (a) + 1)

// Function prototypes
int getGridUsernum(const char *username);
int verifyGridUser(const char *username);
int open_output(const char *outfile);
void exec_useradd(char *DN, const char *caller_username, const char *outfile);
void exec_userdel(char *username);
void exec_getquota(char *username, const char *outfile);
void exec_setquota(char *username, char *quotas[4]);
void privileged_execv(char *const argv[]);
int privileged_kill(uid_t victim_uid, int signal);
void user_wipe(uid_t victim_uid);
static void add_gridmap_entry(const char *DN, const char * grid_username);
static void add_user_to_group( const char * caller_username,
        const char * grid_username);

static
void
remove_gridmap_entry(const char * grid_username);

typedef struct _gridmap_line_s {
  char *dn;
  char **user_ids;
} globus_i_gss_assist_gridmap_line_t;

int
globus_i_gss_assist_gridmap_parse_line(
    char *                              line,
    globus_i_gss_assist_gridmap_line_t **
                                        gline);
void
globus_i_gss_assist_gridmap_line_free(
    globus_i_gss_assist_gridmap_line_t *line);


// Global variables
uid_t real_uid;


/* MAIN
 *
 * Check uid and euid, temporarily drop privileges, parse command line,
 * and invoke appropriate routine.
 */

int main(int argc, char *argv[])
{
    char buff[16], *cp;
    struct stat statbuf;

    if (geteuid() != 0)
    {
        fprintf(stderr, "Sorry, must run with euid root.\n");
        return 1;
    }

    real_uid = getuid();

    // Don't run for arbitrary users -- only for root and globus user
    // Alternatively, we could hard-code an authorized uid/username
    if (real_uid != 0)
    {
        // Get the gid of the file we're running from
        snprintf(buff, 16, "/proc/%d/exe", getpid());
        if (stat(buff, &statbuf) != 0)
        {
            perror("stat");
            return 1;
        }

	/*
        // If user's id doesn't match gid of executable, refuse to run
        if (real_uid != statbuf.st_gid)
        {
            fprintf(stderr,
                    "Sorry, you are not authorized to run this program.\n");
            return 1;
        }
	*/
    }

    // From here we only need root privileges for the exec; drop suid for now
    seteuid(real_uid);

    // Parse arguments, dispatching action to appropriate exec_... function
    if (argc == 4 && strcmp(argv[1], "-adduser") == 0)
    {
        exec_useradd(argv[2], argv[3], NULL);
    }
    else if (argc == 5 && strcmp(argv[1], "-adduser") == 0)
    {
        exec_useradd(argv[2], argv[3], argv[4]);
    }
    else if (argc == 3 && strcmp(argv[1], "-deluser") == 0)
    {
        exec_userdel(argv[2]);
    }
    else if (argc == 3 && strcmp(argv[1], "-getquota") == 0)
    {
        exec_getquota(argv[2], NULL);
    }
    else if (argc == 4 && strcmp(argv[1], "-getquota") == 0)
    {
        exec_getquota(argv[2], argv[3]);
    }
    else if (argc == 7 && strcmp(argv[1], "-setquota") == 0)
    {
        exec_setquota(argv[2], argv + 3);
    }
    else
    {
        cp = strrchr(argv[0], '/');
        fprintf(stderr,
                "Syntax: %s -adduser | -deluser | -getquota | -setquota ...\n",
                cp == NULL ? argv[0] : cp + 1);
        return 1;
    }

    // If exec_... succeeds, we will never reach this point
    fprintf(stderr, "Unexpected error executing specified action.\n");

    return 1;
}


/*
 * FUNCTION getGridUsernum
 *
 * Returns the nonnegative integer embedded in username if username is
 * valid, or -1 otherwise. To be considered valid, username must be the
 * string GRID_USER_PREFIX concatenated with a nonnegative integer, and
 * must be no more than 8 characters in length.  Furthermore, the integer
 * must be between GRID_UNUM_MIN and GRID_UNUM_MAX inclusive.
 */

int getGridUsernum(const char *username)
{
    const char *cp;
    int len, usernum;

    if (username == NULL)
    {
        return -1;
    }

    len = strlen(username);

    if (len <= GRID_USER_PREFIX_LEN || len > 8 ||
        strncmp(username, GRID_USER_PREFIX, GRID_USER_PREFIX_LEN) != 0)
    {
        return -1;
    }

    for (cp = username + GRID_USER_PREFIX_LEN; *cp != '\0'; cp++)
    {
        if (!isdigit(*cp))
        {
            return -1;
        }
    }

    usernum = atoi(username + GRID_USER_PREFIX_LEN);

    if (usernum < GRID_UNUM_MIN || usernum > GRID_UNUM_MAX)
    {
        return -1;
    }

    return usernum;
}


/*
 * FUNCTION verifyGridUser
 *
 * If the specified username corresponds to a valid grid account, return the
 * corresponding UID; otherwise, return -1.
 */

int verifyGridUser(const char *username)
{
    struct passwd *pwd;
    uid_t uid;

    // Verify that username is of the proper format
    if (getGridUsernum(username) < 0)
    {
        fprintf(stderr, "%s not a grid username\n", username);
        return -1;
    }

    // Verify that username exists and that uid is within the allowed range
    if ((pwd = getpwnam(username)) == NULL)
    {
        fprintf(stderr, "username %s not found\n", username);
        return -1;
    }

    uid = pwd->pw_uid;
    endpwent();

    if (uid < GRID_UID_MIN || uid > GRID_UID_MAX)
    {
        fprintf(stderr, "uid %d is outside permitted range\n", uid);
        return -1;
    }

    return uid;
}


/*
 * FUNCTION open_output
 *
 * If outfile is non-NULL, it is safely opened for output and the resulting
 * file descriptor is returned.  If outfile is NULL, the file descriptor for
 * stdout is returned.  Upon error, -1 is returned.
 */

int open_output(const char *outfile)
{
    int fd;
    struct stat statbuf;

    if (!outfile)
    {
        return 1;  // stdout
    }

    if ((fd = open(outfile, O_WRONLY | O_NOFOLLOW)) == -1)
    {
        perror("open");
        return -1;
    }

    if (fstat(fd, &statbuf) != 0)
    {
        perror("fstat");
        close(fd);
        return -1;
    }

    if (!S_ISREG(statbuf.st_mode) ||      // must be a regular file
        statbuf.st_uid != 0 ||     // must be owned by globus user
        //statbuf.st_gid != real_uid ||     // must have group of globus user
        (statbuf.st_mode & S_IWOTH) ||    // must not be writable by others
        statbuf.st_size != 0)             // must currently be empty
    {
        fprintf(stderr, "Unsafe output file\n");
        close(fd);
        return -1;
    }

    return fd;
}


/*
 * PROCEDURE exec_useradd
 *
 * Create a UNIX account for grid user DN, as follows:
 *
 * 1) Determine the uid and username of the last grid user account created.
 *    The code attempts to read this info from the lockfile, but if it's
 *    not found, it goes through /etc/passwd to find the last account
 *    created and still in use (if there is one!).
 *
 * 2) Determine the next AVAILABLE uid and username within the #define'd
 *    limits, wrapping around to ..._MIN if necessary.
 *
 * 3) Write chosen username and uid to file/stdout, communicating them to
 *    the Java grid service.
 *  **CHANGED: prints username:group
 *
 * 4) Build command line and execute useradd AS ROOT.
 *
 * Upon success, does not return.
 */

void exec_useradd(char *DN, const char * caller_username, const char *outfile)
{
    const struct passwd *pwd;
    char grid_homedir[15] = "/home/";       // Change for different home dir
    char grid_username[9], grid_uid_str[10];
    uid_t grid_uid = GRID_UID_MIN;
    int output_fd, lock_fd, i, grid_unum = -1;
    int digitsRead, state;
    char *args[14];
    pid_t pid;

    // Open the output.
    if ((output_fd = open_output(outfile)) == -1) {
        return;
    }

    // Use advisory locking to prevent a race condition below, in case
    // multiple copies of this program are running simultaneously
    seteuid(0);
    lock_fd = open(INFO_LOCKFILE,
                   O_CREAT | O_RDWR | O_NOFOLLOW, S_IRUSR | S_IWUSR);
    if (lock_fd == -1) {
        perror("open");
        return;
    }
    flock(lock_fd, LOCK_EX);
    seteuid(real_uid);

    // The lockfile should contain a string "unum:uid;" indicating the last
    // grid usernum and uid we attempted to assign.  Read in this info.
    digitsRead = 0;
    state = 0;
    while (read(lock_fd, grid_uid_str + digitsRead, 1) > 0)
    {
        if (isdigit(grid_uid_str[digitsRead]) && digitsRead < 9)
        {
            digitsRead++;
        }
        else if (grid_uid_str[digitsRead] == ':' && state == 0)
        {
            grid_uid_str[digitsRead] = '\0';
            grid_unum = atoi(grid_uid_str);
            digitsRead = 0;
            state = 1;
        }
        else if (grid_uid_str[digitsRead] == ';' && state == 1)
        {
            grid_uid_str[digitsRead] = '\0';
            grid_uid = atoi(grid_uid_str);
            digitsRead = 0;
            state = 2;
            break;
        }
        else
        {
            break;  // invalid input
        }
    }

    // If we didn't read anything useful from the lockfile, try passwd file.
    if (state != 2 || grid_unum < GRID_UNUM_MIN || grid_unum > GRID_UNUM_MAX ||
        grid_uid < GRID_UID_MIN || grid_uid > GRID_UID_MAX)
    {
        grid_unum = -1;

        // Get uid and username of last grid user account in /etc/passwd
        setpwent();
        while ((pwd = getpwent()))
        {
            if (pwd->pw_uid >= GRID_UID_MIN &&
                pwd->pw_uid <= GRID_UID_MAX &&
                (i = getGridUsernum(pwd->pw_name)) >= 0)
            {
                grid_uid = pwd->pw_uid;
                grid_unum = i;
            }
        }
    }

    // If no grid accounts were found, start with ..._MIN, else increment
    if (grid_unum == -1)
    {
        grid_uid = GRID_UID_MIN;
        grid_unum = GRID_UNUM_MIN;
    }
    else
    {
        grid_uid = NEXT_GRID_UID(grid_uid);
        grid_unum = NEXT_GRID_UNUM(grid_unum);
    }

    // Get next FREE available grid user uid (if there is one!)
    for (i = GRID_UID_MIN; i <= GRID_UID_MAX; i++)
    {
        if (getpwuid(grid_uid) == NULL)
        {
            break;
        }
        else
        {
            grid_uid = NEXT_GRID_UID(grid_uid);
        }
    }
    if (i > GRID_UID_MAX)
    {
        fprintf(stderr, "No available UIDs.\n");
        return;
    }

    // Get next FREE available grid username (if there is one!)
    for (i = GRID_UNUM_MIN; i <= GRID_UNUM_MAX; i++)
    {
        sprintf(grid_username, GRID_UNAME_FORMAT, GRID_USER_PREFIX, grid_unum);
        if (getpwnam(grid_username) == NULL)
        {
            break;
        }
        else
        {
            grid_unum = NEXT_GRID_UNUM(grid_unum);
        }
    }
    if (i > GRID_UNUM_MAX)
    {
        fprintf(stderr, "No available usernames.\n");
        return;
    }

    endpwent();

    // We now know what username and uid we're assigning.
    sprintf(grid_uid_str, "%d", grid_uid);

    // Communicate the chosen username to the Java caller
    if (write(output_fd, grid_username, strlen(grid_username)) == -1 ||
        write(output_fd, ":", 1) == -1 || 
        write(output_fd, GRID_USER_GROUP, strlen(GRID_USER_GROUP)) == -1 ||
        write(output_fd, "\n", 1) == -1) 
    {
        perror("write");
        return;
    }
    close(output_fd);

    // Record the chosen username and uid in the lockfile.
    lseek(lock_fd, 0, SEEK_SET);
    write(lock_fd, grid_username + GRID_USER_PREFIX_LEN,
          strlen(grid_username) - GRID_USER_PREFIX_LEN);
    write(lock_fd, ":", 1);
    write(lock_fd, grid_uid_str, strlen(grid_uid_str));
    write(lock_fd, ";\n", 2);

    // Close the lockfile (implicitly releases the lock).
    close(lock_fd);

    // DN must not have colons in it! (allowed?) -- but useradd checks for
    // this
    //
    // Modified 2004-08-13 to remove -n (create group with same name as
    // the new username) and have the GRID_USER_GROUP be a supplemental
    // group instead of the default group
    args[0] = USERADD_PROG;
    args[1] = "-c";
    args[2] = DN;
    args[3] = "-d";
    args[4] = strcat(grid_homedir, grid_username);
    args[5] = "-G";
    args[6] = GRID_USER_GROUP;
    args[7] = "-m";
    args[8] = "-s";
    args[9] = GRID_USER_SHELL;
    args[10] = "-u";
    args[11] = grid_uid_str;
    args[12] = grid_username;
    args[13] = NULL;

    // Log new account-DN mapping to /var/log/secure
    openlog("sandbox_setuid", LOG_PID, LOG_AUTHPRIV);
    syslog(LOG_INFO, "creating account `%s' for %s", grid_username, DN);
    closelog();

    // Update grid-mapfile to contain mapping for the DN to the new username
    add_gridmap_entry(DN, grid_username);

    // create user
    if ((pid = fork()) == 0)
    {
        while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
    }
    else if (pid > 0)
    {
	privileged_execv(args);
    }
    else
    {
	perror("fork");
        exit(1);
    }
    

    // modify calling user to be in new group
    add_user_to_group(caller_username, grid_username);

    // if exec fails, remove the grid map entry
    remove_gridmap_entry(grid_username);
}


/*
 * PROCEDURE exec_userdel
 *
 * Delete a UNIX account for grid user "username" by executing userdel AS ROOT.
 * Any processes owned by the user are killed and the user's home directory is
 * removed, as well as any of the user's files in /tmp.
 *
 * Upon success, does not return.
 */

void exec_userdel(char *username)
{
    char *args[4];
    int uid;

    if ((uid = verifyGridUser(username)) == -1)
    {
        return;
    }

    // Kill any of the user's processes still running
    if (privileged_kill(uid, SIGTERM))
    {
        sleep(1);
        if (privileged_kill(uid, SIGTERM))
        {
            sleep(1);
            if (privileged_kill(uid, SIGKILL))
            {
                sleep(1);
                privileged_kill(uid, SIGKILL);
            }
        }
    }

    // Wipe out the user's files
    user_wipe(uid);

    args[0] = USERDEL_PROG;
    args[1] = "-r";
    args[2] = username;
    args[3] = NULL;

    remove_gridmap_entry(username);
    privileged_execv(args);
}


/*
 * PROCEDURE exec_getquota
 *
 * Get the quota on a UNIX account for grid user "username" by executing
 * quota AS ROOT.
 *
 * Upon success, does not return.
 */

void exec_getquota(char *username, const char *outfile)
{
    char *args[4];
    int output_fd;

    if (verifyGridUser(username) == -1)
    {
        return;
    }

    // Open the output.
    if ((output_fd = open_output(outfile)) == -1)
    {
        return;
    }

    dup2(output_fd, 1);

    args[0] = QUOTA_PROG;
    args[1] = "-v";
    args[2] = username;
    args[3] = NULL;

    privileged_execv(args);
}


/*
 * PROCEDURE exec_setquota
 *
 * Set the quota on a UNIX account for grid user "username" by executing
 * setquota AS ROOT.
 *
 * The four quota parameters: block-soft, block-hard, inode-soft, inode-hard
 *
 * Upon success, does not return.
 */

void exec_setquota(char *username, char *quotas[4])
{
    char *cp, *args[8];
    int i;

    if (verifyGridUser(username) == -1)
    {
        return;
    }

    // Make sure that quotas all are numeric and nonnegative
    for (i = 0; i < 4; i++)
    {
        for (cp = quotas[i]; *cp != '\0'; cp++)
        {
            if (!isdigit(*cp))
            {
                fprintf(stderr, "Invalid quota specified: %s\n", quotas[i]);
                return;
            }
        }
    }

    args[0] = SETQUOTA_PROG;
    args[1] = username;
    args[2] = quotas[0];
    args[3] = quotas[1];
    args[4] = quotas[2];
    args[5] = quotas[3];
    args[6] = "-a";
    args[7] = NULL;

    privileged_execv(args);
}


/*
 * PROCEDURE privileged_execv
 *
 * Calls execv() AS ROOT, dropping all privileges and returning upon failure.
 */

void privileged_execv(char *const argv[])
{
    seteuid(0);
    execv(argv[0], argv);
    perror("execv");
    setuid(real_uid);
    return;
}


/*
 * FUNCTION privileged_kill
 *
 * If signal > 0, sends specified signal to all of a user's processes.
 * Returns the number of user's processes found, or -1 upon error.
 */

int privileged_kill(uid_t victim_uid, int signal)
{
    DIR *procdir;
    struct dirent *de;
    struct stat statbuf;
    int pid, num_procs = 0;
    char path[17] = "/proc", *file = path + 5;

    // Just a precaution... don't kill all of root's processes!!!
    if (signal && !victim_uid)
    {
        return -1;
    }

    if (!(procdir = opendir(path)))
    {
        perror("opendir");
        return -1;
    }

    *file = '/';
    ++file;

    // We only need privileges if we may be sending signal(s)
    if (signal > 0)
    {
        seteuid(0);
    }

    while ((de = readdir(procdir)))
    {
        if ((pid = atoi(de->d_name)))
        {
            strncpy(file, de->d_name, 10);
            file[10] = '\0';
            if (stat(path, &statbuf) == 0 && statbuf.st_uid == victim_uid)
            {
                ++num_procs;
                if (signal > 0)
                {
                    kill(pid, signal);
                }
            }
        }
    }

    if (signal > 0)
    {
        seteuid(real_uid);
    }

    closedir(procdir);

    return num_procs;
}


/*
 * PROCEDURE user_wipe
 *
 * Wipes all of the user's files and directories from /tmp and home directory.
 */

void user_wipe(uid_t victim_uid)
{
    char *args[15], grid_uid_str[10];
    pid_t pid;
    struct passwd *pwd;

    // Get the user's home directory.
    if ((pwd = getpwuid(victim_uid)) == NULL)
    {
        fprintf(stderr, "uid %d not found\n", victim_uid);
        return;
    }

    if (!pwd->pw_dir)
    {
        return;
    }

    // Delete all files owned by user
    sprintf(grid_uid_str, "%d", victim_uid);
    args[0] = FIND_PROG;
    args[1] = pwd->pw_dir;
    args[2] = "/tmp";
    args[3] = "-xdev";
    args[4] = "-uid";
    args[5] = grid_uid_str;
    args[6] = "!";
    args[7] = "-type";
    args[8] = "d";
    args[9] = "-exec";
    args[10] = RM_PROG;
    args[11] = "-f";
    args[12] = "{}";
    args[13] = ";";
    args[14] = NULL;

    if ((pid = fork()) == -1)
    {
        perror("fork");
    }
    else if (pid)
    {
        while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
    }
    else
    {
        seteuid(0);
        setuid(victim_uid);
        close(2);
        open("/dev/null", O_WRONLY);
        execv(args[0], args);
        perror("execv");
        exit(1);
    }

    // Remove user directories
    args[0] = FIND_PROG;
    args[1] = pwd->pw_dir;
    args[2] = "/tmp";
    args[3] = "-xdev";
    args[4] = "-depth";
    args[5] = "-uid";
    args[6] = grid_uid_str;
    args[7] = "-type";
    args[8] = "d";
    args[9] = "-exec";
    args[10] = RMDIR_PROG;
    args[11] = "{}";
    args[12] = ";";
    args[13] = NULL;

    if ((pid = fork()) == -1)
    {
        perror("fork");
    }
    else if (pid)
    {
        while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
    }
    else
    {
        seteuid(0);
        setuid(victim_uid);
        close(2);
        open("/dev/null", O_WRONLY);
        execv(args[0], args);
        perror("execv");
        exit(1);
    }

    return;
}

/*
 * PROCEDURE add_gridmap_entry
 *
 * Atomically create a new entry in the grid mapfile mapping the DN
 * to the specified username.
 */
static
void
add_gridmap_entry(const char *DN, const char * grid_username)
{
    int gridmap_fd;
    char * gridmap_buffer;
    int rc;
    struct stat stat;
    size_t len;
    size_t off;
    off_t truncate_len;
    int didit=0;
    globus_i_gss_assist_gridmap_line_t * gline;

    seteuid(0);
    gridmap_fd = open(GRIDMAP, O_RDWR|O_CREAT);

    if (gridmap_fd < 0)
    {
        perror("open("GRIDMAP")");
        exit(1);
    }

    rc = flock(gridmap_fd, LOCK_EX);
    if (rc < 0)
    {
        perror("flock(gridmap)");
        exit(1);
    }

    rc = fstat(gridmap_fd, &stat);
    if (rc < 0)
    {
        perror("fstat(gridmap)");
        exit(1);
    }

    len = (size_t)stat.st_size;

    gridmap_buffer = malloc(len);
    if (gridmap_buffer == NULL)
    {
        perror("malloc(gridmap_buffer)");
        exit(1);
    }
    rc = read(gridmap_fd, gridmap_buffer, len);
    if (rc < len)
    {
        perror("read("GRIDMAP")");
        exit(1);
    }
    off = 0;
    lseek(gridmap_fd, 0, SEEK_SET);
    while (off < len)
    {
        char * eol = memchr(gridmap_buffer + off, '\n', off-len);
        if (eol == NULL)
        {
            char * tmp = realloc(gridmap_buffer, len+1);
            if (tmp == NULL)
            {
                perror("realloc");
                exit(1);
            }
            gridmap_buffer = tmp;
            gridmap_buffer[len++] = '\n';
            eol = gridmap_buffer + len;
        }

        rc = globus_i_gss_assist_gridmap_parse_line(
                gridmap_buffer+off,
                &gline);
        if (rc != 0 || gline == NULL)
        {
            fprintf(stderr,
                    "ERROR PARSING GRIDMAP LINE: %s",
                    gridmap_buffer+off);
            exit(1);
        }
        write(gridmap_fd, "\"", 1);
        write(gridmap_fd, gline->dn, strlen(gline->dn));
        write(gridmap_fd, "\" ", 2);

        if (strcmp(DN, gline->dn) == 0)
        {
            char ** user_ids;

            user_ids = gline->user_ids;

            while (*user_ids != NULL)
            {
                write(gridmap_fd, *user_ids, strlen(*user_ids));
                user_ids++;
                write(gridmap_fd, ",", 1);
            }

            write(gridmap_fd, grid_username, strlen(grid_username));

            didit=1;
        }
        else
        {
            char ** user_ids = gline->user_ids;

            while (*user_ids != NULL)
            {
                write(gridmap_fd, *user_ids, strlen(*user_ids));
                user_ids++;
                if (*user_ids)
                {
                    write(gridmap_fd, ",", 1);
                }
            }
        }
        globus_i_gss_assist_gridmap_line_free(gline);

        off = eol - gridmap_buffer + 1;
        write(gridmap_fd, "\n", 1);
    }

    // DN not found in gridmap
    if (!didit)
    {
        write(gridmap_fd, "\"", 1);
        write(gridmap_fd, DN, strlen(DN));
        write(gridmap_fd, "\" ", 2);
        write(gridmap_fd, grid_username, strlen(grid_username));
        write(gridmap_fd, "\n", 1);
    }
    truncate_len = lseek(gridmap_fd, 0, SEEK_CUR);
    ftruncate(gridmap_fd, truncate_len);
    close(gridmap_fd);
    seteuid(real_uid);
}

static
void
remove_gridmap_entry(const char * grid_username)
{
    int gridmap_fd;
    char * gridmap_buffer;
    int rc;
    struct stat stat;
    size_t len;
    size_t off;
    globus_i_gss_assist_gridmap_line_t * gline;
    off_t truncate_len;
    seteuid(0);

    gridmap_fd = open(GRIDMAP, O_RDWR|O_CREAT);

    if (gridmap_fd < 0)
    {
        perror("open("GRIDMAP")");
        exit(1);
    }

    rc = flock(gridmap_fd, LOCK_EX);
    if (rc < 0)
    {
        perror("flock(gridmap)");
        exit(1);
    }

    rc = fstat(gridmap_fd, &stat);
    if (rc < 0)
    {
        perror("fstat(gridmap)");
        exit(1);
    }

    len = (size_t)stat.st_size;

    gridmap_buffer = malloc(len);
    if (gridmap_buffer == NULL)
    {
        perror("malloc(gridmap_buffer)");
        exit(1);
    }
    rc = read(gridmap_fd, gridmap_buffer, len);
    if (rc < len)
    {
        perror("read("GRIDMAP")");
        exit(1);
    }
    off = 0;
    lseek(gridmap_fd, 0, SEEK_SET);
    while (off < len)
    {
        char * eol = memchr(gridmap_buffer + off, '\n', off-len);
        char ** user_ids;

        if (eol == NULL)
        {
            char * tmp = realloc(gridmap_buffer, len+1);
            if (tmp == NULL)
            {
                perror("realloc");
                exit(1);
            }
            gridmap_buffer = tmp;
            gridmap_buffer[len++] = '\n';
            eol = gridmap_buffer + len;
        }

        rc = globus_i_gss_assist_gridmap_parse_line(
                gridmap_buffer+off,
                &gline);
        if (rc != 0 || gline == NULL)
        {
            fprintf(stderr,
                    "ERROR PARSING GRIDMAP LINE: %s",
                    gridmap_buffer+off);
            exit(1);
        }


        user_ids = gline->user_ids;
        /* strip out matching username if present */
        while (*user_ids != NULL)
        {
            if (strcmp(*user_ids, grid_username) == 0)
            {
                char ** tmp = user_ids;

                free(*tmp);
                while(*(tmp)) {
                    *tmp = *(tmp+1);
                    tmp++;
                }

            }
            user_ids++;
        }

        if (*gline->user_ids != NULL)
        {
            /* there are remaining mappings for this DN */
            write(gridmap_fd, "\"", 1);
            write(gridmap_fd, gline->dn, strlen(gline->dn));
            write(gridmap_fd, "\" ", 2);

            user_ids = gline->user_ids;

            while (*user_ids != NULL)
            {
                write(gridmap_fd, *user_ids, strlen(*user_ids));
                if (*(++user_ids) != NULL)
                {
                    write(gridmap_fd, ",", 1);
                }
            }
            write(gridmap_fd, "\n", 1);
        }

        globus_i_gss_assist_gridmap_line_free(gline);
        off = eol - gridmap_buffer + 1;
    }
    truncate_len = lseek(gridmap_fd, 0, SEEK_CUR);
    ftruncate(gridmap_fd, truncate_len);
    close(gridmap_fd);
    seteuid(real_uid);
}

/*
static
int
mygetgrouplist(const char * username, gid_t gid, gid_t groups[], int * ng)
{
    struct group * group;
    int i;
    int max = *ng;

    *ng = 0;

    setgrent();

    if (i < max)
    {
        groups[i++] = gid;
    }
    (*ng)++;
    while ((group = getgrent()) != NULL)
    {
        int j;

        for (j = 0; group->gr_mem[j] != NULL; j++)
        {
            if(strcmp(group->gr_mem[j], username) == 0)
            {
                (*ng)++;
                if (i < max)
                {
                    groups[i++] = gid;
                }
                break;
            }
        }
    }

    if ((*ng) > max)
    {
        errno = EFAULT;
        return -1;
    }
    else
    {
        return 0;
    }
}
*/
static
void
add_user_to_group(
    const char * caller_username,
    const char * grid_username)
{
    struct passwd * pw;
    struct group * gr;
    int ng = 256;
    gid_t groups[256];
    char * args[5];
    int rc;
    int i;

    pw = getpwnam(caller_username);

    if (pw == NULL)
    {
        perror("getpwnam");
    }

    endgrent();
    sleep(2);
    gr = getgrnam(grid_username);
    if (gr == NULL)
    {
        perror("getgrnam");
    }
    rc = getgrouplist(caller_username, gr->gr_gid, groups, &ng);
    if (rc < 0)
    {
        perror("getgrouplist");
	exit(2);
    }


    args[0] = USERMOD_PROG;
    args[1] = "-G";
    args[2] = malloc(ng * 8);
    args[2][0] = '\0';

    for (i = 0; i < ng; i++)
    {
        char groupstr[7];

        if (groups[i] == pw->pw_gid)
        {
            continue;
        }
        sprintf(groupstr, "%u", (unsigned int) groups[i]);
        strcat(args[2], groupstr);
        if (i < (ng-1))
        {
            strcat(args[2], ",");
        }
    }
    args[3] = (char *) caller_username;
    args[4] = NULL;

    privileged_execv(args);
}


