/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */

/**
 * 
 * Routines to map a globusID to a local unix user
 *
 **/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
 
#include "log.h"
#include "util.h"
#include "errors.h"
#include "gridmap.h"

typedef struct _gridmap_line_s {
  char *dn;
  char **user_ids;
} gridmap_line_t;


#define WHITESPACE_CHARS		" \t\n"
#define QUOTING_CHARS			"\""
#define ESCAPING_CHARS			"\\"
#define COMMENT_CHARS			"#"
/* Characters seperating user ids in the gridmap file */
#define USERID_SEP_CHARS		","
/*
 * Characters that terminate a user id in the gridmap file. This
 * is a combination of whitespace and seperators.
 */
#define USERID_TERMINATOR_CHARS		USERID_SEP_CHARS WHITESPACE_CHARS

#ifndef NUL
#define NUL				'\0'
#endif
/*
 * Number of user id slots to allocate at a time
 * Arbitraty value, but must be >= 2.
 */
#define USERID_CHUNK_SIZE		4

static int gridmap_find_dn(
    const char * const 	 dn,
    gridmap_line_t **		 gline);
    
static int gridmap_parse_line(
    char * 			        line,
    gridmap_line_t **	  gline);
    
static int gridmap_parse_globusid(
    const char *  unparsed,
    char **				pparsed);
    
void gridmap_line_free(gridmap_line_t * gline);

static int xdigit_to_value(char xdigit);
 
/*
 * REQUIRED: GRIDMAP environment variable pointing at the
 * map file.
 *
 * @param globusidp
 *        the GSSAPI name from the client who requested
 *        authentication
 * @param useridp
 *        the resulting user ID name for the local system
 *
 * @return 
 *        0 on success
 *        -1 if bad arguments
 *        1 on error
 */
extern int assist_gridmap(
    char *  globusidp, 
    char ** useridp) 
{
  char *func = "assist_gridmap()";
  gridmap_line_t *gline = NULL;
  int ret = 0;
  
  debug_msg(8,"%s",func);
  
  /* Check arguments */
  if (globusidp == NULL) {
   log_msg("%s problem, globusidp is null",func);
   fprintf(stderr,"%s problem, globusidp is null",func);
   return GENERAL_FAILURE;
  } 
  
  if (useridp == NULL) {
   log_msg("%s problem, useridp is null",func);
   fprintf(stderr,"%s problem, useridp is null",func);
   return GENERAL_FAILURE;
  }

  *useridp = NULL;
  
  ret = gridmap_find_dn(globusidp, &gline);
  if (ret != SUCCESS) { 
    log_msg("%s problem with gridmap_find_dn",func);
    fprintf(stderr,"%s problem with gridmap_find_dn",func);
    return ENVIRONMENT;
  }
  
  
  if (gline != NULL)
  {
    if ((gline->user_ids == NULL) ||
        (gline->user_ids[0] == NULL))
    {
	    /*
	     * If we get here then something in this code is broken
	     * or the gridmap file is badly formatted or, most likely,
	     * both.
	     */
       log_msg("%s problem, Invalid (NULL) user id values\n",func);
       fprintf(stderr,"%s Invalid (NULL) user id values\n",func);
       return GENERAL_FAILURE;
    }
  
    /* First user id is default */
    *useridp = safe_dup(gline->user_ids[0]);
    
    gridmap_line_free(gline);
    
    if (*useridp == NULL) {
      log_msg("%s problem, Duplicate string operation failed\n",func);
      fprintf(stderr,"%s problem, Duplicate string operation failed\n",func);
      return GENERAL_FAILURE;
    }
  
  } else {
  	/* No entry found in gridmap file for this user */
    log_msg("The DN: %s could not be mapped to a valid user in the "
             "gridmap file\n",globusidp);
    return NO_USER;
  }
  
  return 0;
}

/**
 * Locate the entry for the given DN in the default gridmap file
 */

static int gridmap_find_dn(
    const char * const 	 dn,
    gridmap_line_t **		 gline)
{
  char *            gridmap_filename = NULL;
  char *	          open_mode = "r";
  FILE *		        gmap_stream = NULL;
  int				        found = 0;
  int               ret = 0;
  gridmap_line_t *  gline_tmp = NULL;
  struct stat       statbuf;
  
    
  char *func = "gridmap_find_dn()";
  
  debug_msg(8,"%s",func);
  
  
  /* Check arguments */
  if (dn == NULL) {
   log_msg("%s problem, dn is null",func);
   fprintf(stderr,"%s problem, dn is null",func);
   return GENERAL_FAILURE;
  }
  
  if (gline == NULL) {
   log_msg("%s problem, gline is null",func);
   fprintf(stderr,"%s problem, gline is null",func);
   return GENERAL_FAILURE;
  }
  
  if (getenv("GRIDMAP") == NULL) {
    log_msg("%s problem: GRIDMAP not in environment",func);
    fprintf(stderr,"%s problem: GRIDMAP not in environment",func);
    return GENERAL_FAILURE; /* GRIDMAP defined? */
  }
  
  gridmap_filename = safe_dup(getenv("GRIDMAP"));
  
  ret = stat(gridmap_filename, &statbuf);
  if (ret) {
    log_msg("grid-mapfile (%s) does not exist\n", gridmap_filename);
    fprintf(stderr, "grid-mapfile (%s) does not exist\n", gridmap_filename);
    return ENVIRONMENT;
  }
  
  gmap_stream = fopen(gridmap_filename, open_mode);
  
  if (gmap_stream == NULL) {
    log_msg("Couldn't open gridmap file: %s for reading\n", gridmap_filename);
    fprintf(stderr,"Couldn't open gridmap file: %s for reading\n", gridmap_filename);
    return ENVIRONMENT;
  }

  free(gridmap_filename);
  gridmap_filename = NULL;
  
  do {
    char 				line[1024];

    if (fgets(line, sizeof(line), gmap_stream) == NULL)
    {
	    break;		/* EOF or error */
    }

    ret = gridmap_parse_line(line, &gline_tmp);
    if (ret != SUCCESS)
    {
      debug_msg(3,"%s problem: gridmap_parse_line (line %s) returns %d",func, line, ret);
      continue;		/* Parse error */
    }

    if ((gline_tmp != NULL) && (strcmp(dn, gline_tmp->dn) == 0))
    {
      found = 1;
    }
    else
    {
	    gridmap_line_free(gline_tmp);
    }

  } while (!found);

  fclose(gmap_stream);
  gmap_stream = NULL;

  if (found)
  {
      *gline = gline_tmp;
      
  }
  else
  {
      *gline = NULL;
  }
    
  return ret;
  
}

/**
 * 
 * Given a line from the gridmap file, parse it returning
 * a gridmap_line_t structure. line is modified during parsing.
 * The format of the line is expected to be:
 * <DN> <userid>[,<userid>[,<userid>...]]
 * Leading and trailing whitespace is ignored.
 * userids must only have a comma between them, no whitespace.
 * Anything after the userids is ignored.
 * Anything after an unescaped comment character is ignored.
 *
 * @param line
 *        the line to parse
 * @param gline
 *        the resulting parsed gridmap line structure
 *
 */
static int gridmap_parse_line(
    char * 			        line,
    gridmap_line_t **	  gline)
{
    char *				    dn_end;
    char *				    parsed_dn = NULL;
    char **				    userids = NULL;
    int					      num_userids = 0;
    int					      userid_slots = 0;
    gridmap_line_t *  gline_tmp = NULL;
    int               result = SUCCESS;
  
    char *func = "gridmap_parse_line()";
    
    debug_msg(8,"%s",func);
    
    /* Check arguments */
    if (line == NULL)
    {
         log_msg("%s problem, line is null",func);
         fprintf(stderr,"%s problem, line is null",func);
         result = GENERAL_FAILURE;
         goto exit;
    }
    if (gline == NULL)
    {
         log_msg("%s problem, gline is null",func);
         fprintf(stderr,"%s problem, gline is null",func);
         result = GENERAL_FAILURE;
         goto exit;
    }
    
    debug_msg(7,"%s, line: %s",func,line);
    

    /* Skip over leading whitespace */
    line += strspn(line, WHITESPACE_CHARS);

    /* Check for comment at start of line and ignore line if present */
    if (strchr(COMMENT_CHARS, *line) != NULL) 
    {
        /* Ignore line, return NULL gline */
        *gline = NULL;
        goto exit;
    }
	
    /* Check for empty line */
    if (*line == NUL)
    {
        /* Empty line, return NULL gline. */
        *gline = NULL;
        goto exit;
    }

    /* Is DN quoted? */
    if (strchr(QUOTING_CHARS, *line) != NULL)
    {
	/*
	 * Yes, skip over opening quote and look for unescaped
	 * closing double quote
	 */
	line++;
	dn_end = line;

	do
	{
	    dn_end += strcspn(dn_end, QUOTING_CHARS);

	    if (*dn_end == NUL)
      {
          debug_msg(3, "A closing quote is missing in the gridmap file, on the line:\n%s\n", line);
          result = ENVIRONMENT;
          goto exit;
      }

	    /* Make sure it's not escaped */
	}
	while (strchr(ESCAPING_CHARS, *(dn_end - 1)) != NULL);
    }
    else
    {
	/* No, just find next whitespace */
	dn_end = line + strcspn(line, WHITESPACE_CHARS);

	if (*dn_end == NUL)
        {
          debug_msg(1, "A closing quote is missing in the gridmap file, on the line:\n%s\n", line);
          result = ENVIRONMENT;
          goto exit;
        }
    }

    /* NUL terminate DN and parse */
    *dn_end = NUL;

    result = gridmap_parse_globusid(line, &parsed_dn);
    if(result != SUCCESS)
    {
        debug_msg(1, "gridmap_parse_globusid != 0");
        goto exit;
    }

    /* Skip over closing delim and any whitespace after DN */
    line = dn_end + 1;
    line += strspn(line, WHITESPACE_CHARS);

    /* Parse list of unix user ID seperated by USERID_SEP_CHARS */
    while ((*line != NUL) &&
	   (strchr(WHITESPACE_CHARS, *line) == NULL))
    {
	int					userid_len;

	/* Find end of this userid */
	userid_len = strcspn(line, USERID_TERMINATOR_CHARS);

	/* Make sure we have a slot and if not allocate it */
	if ((num_userids + 1 /* new entry */+ 1 /* for NULL */) > userid_slots)
	{
	    char **userids_tmp;
	    userid_slots += USERID_CHUNK_SIZE;
	    userids_tmp = realloc(userids, userid_slots * sizeof(char *));

	    if (!userids_tmp)
            {
                result = GENERAL_FAILURE;
                debug_msg(1,"Could not allocate enough memory");
                goto error_exit;
            }

	    userids = userids_tmp;
	}
  
	userids[num_userids] = malloc(userid_len + 1 /* for NUL */);

	if (!userids[num_userids])
        {
                result = GENERAL_FAILURE;
                debug_msg(1,"Could not allocate enough memory");
                goto error_exit;
        }

	strncpy(userids[num_userids], line, userid_len);
	userids[num_userids][userid_len] = NUL;

	num_userids++;
	userids[num_userids] = NULL;

	line += userid_len;

	/* If we're on a seperator character, skip over it */
	if (strchr(USERID_SEP_CHARS, *line) != NULL)
        {
	    line++;
        }
    }

    /*
     * There might be more stuff on the line, but we're ignoring
     * it for now.
     */

    /* Ok, build our gridmap_line_t structure */
    gline_tmp = malloc(sizeof(*gline_tmp));

    if (gline_tmp == NULL)
    {
                result = GENERAL_FAILURE;
                debug_msg(1,"Could not allocate enough memory");
                goto error_exit;
    }

    gline_tmp->dn = parsed_dn;
    gline_tmp->user_ids = userids;

    *gline = gline_tmp;
  
    goto exit;

 error_exit:

    if (parsed_dn)
    {
        free(parsed_dn);
    }

    if (userids)
    {
	char **userids_tmp = userids;

	while (*userids_tmp != NULL)
        {
            free(*userids_tmp++);
        }

	free(userids);
    }

    if (gline_tmp)
    {
        free(gline_tmp);
    }

 exit:
    return result;
} 

/**
 * Given a pointer to a string containing the globusid from the
 * gridmap file, return a pointer to a string containing the
 * parsed from of the id.
 *
 * Specifically handle backslashed characters - e.g. '\\',
 * '\x4a' or '\37'.
 *
 * @param unparsed
 *        the unparsed globusid
 * @param pparsed
 *        the resulting parsed string - this should be freed when
 *        no longer needed
 * @result
 *        0 on success, otherwise an error
 **/
static int gridmap_parse_globusid(
    const char *  unparsed,
    char **				pparsed)
{
    /* Is the current character escaped? (Previous char was backslash) */
    int					escaped = 0;

  /* Buffer we are putting resulting name into */
    char *				buffer = NULL;

    /* Buffer's length in bytes */
    int					buffer_len = 0;

    /* And our current position in buffer */
    int					buffer_index = 0;

    /* Character we're currently looking at */
    char			        unparsed_char;

    int  result = SUCCESS;
    
      
    char *func = "gridmap_parse_globusid()";
    
    debug_msg(8,"%s",func);

    /*
     * Check input parameters for legality
     */
    if ((unparsed == NULL) ||
	(pparsed == NULL))
    {
         log_msg("%s problem, line is null",func);
         fprintf(stderr,"%s problem, line is null",func);
         result = GENERAL_FAILURE;
         goto exit;
    }

    buffer_len = strlen(unparsed);
    buffer = malloc(buffer_len);

    if (buffer == NULL)
    {
                result = GENERAL_FAILURE;
                debug_msg(1,"%s Could not allocate enough memory", func);
                goto exit;
    }

    /*
     * Walk through the name, parsing as we go
     */
    while ((unparsed_char = *(unparsed++)) != NUL)
    {
	/* Unescaped backslash */
	if (strchr(ESCAPING_CHARS, unparsed_char) && !escaped)
	{
	    escaped = 1;
	    continue;
	}

	/* Escaped hex character - e.g. '\xfe' */
	if ((unparsed_char == 'x') && escaped)
	{
	    if (isxdigit(*unparsed) &&
		isxdigit(*(unparsed + 1)))
	    {
		/* Set unparsed_char to value represented by hex value */
		unparsed_char =
		    (xdigit_to_value(*unparsed) << 4) +
		    xdigit_to_value(*(unparsed + 1));
	
		unparsed += 2;
	    }
	    /* else just fall through */
	}

	/*
         * Ok, we now have the character in unparsed_char to be appended
         * to our output string.
         *
         * First, make sure we have enough room in our output buffer.
         */

	if ((buffer_index + 1 /* for NUL */) >= buffer_len)
	{
	    /* Grow buffer */
	    char *tmp_buffer;

	    buffer_len *= 2;

	    tmp_buffer = realloc(buffer, buffer_len);

	    if (tmp_buffer == NULL)
	    {
          free(buffer);
          result = GENERAL_FAILURE;
          debug_msg(1,"%s Could not allocate enough memory", func);
          goto exit;
	    }
            
	    buffer = tmp_buffer;
	}
        
	buffer[buffer_index++] = unparsed_char;
	buffer[buffer_index] = NUL;

	escaped = 0;
    }
    
    /* XXX What if escaped == 1 here? */
    /* Success */
    
    *pparsed = buffer;
    
 exit:
    return result;
}


/**
 * Frees all memory allocated to a gridmap_line_t structure.
 *
 * @param gline
 *        pointer to structure to be freed.
 * 
 * @return
 *        void
 */
void gridmap_line_free(gridmap_line_t * gline)
{
    if (gline != NULL)
    {
        if (gline->dn != NULL)
        {
            free(gline->dn);
        }
        
        if (gline->user_ids != NULL)
        {
            char ** userids_tmp = gline->user_ids;
            
            while (*userids_tmp != NULL)
            {
                free(*userids_tmp++);
            }
            
            free(gline->user_ids);
        }
        
        free(gline);
    }
} 


/**
 * Convert an ascii character representing a hexadecimal digit
 * into an integer.
 *
 * @param xdigit
 *        character contianing the hexidecimal digit
 *
 * @return
 *        the value in the xdigit, or -1 if error
 */
static int xdigit_to_value(
    char 				xdigit)
{
    if ((xdigit >= '0') && (xdigit <= '9'))
	return (xdigit - '0');

    if ((xdigit >= 'a') && (xdigit <= 'f'))
	return (xdigit - 'a' + 0xa);

    if ((xdigit >= 'A') && (xdigit <= 'F'))
	return (xdigit - 'A' + 0xa);

    /* Illegal digit */
    return -1;
} 

