/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */

/*
 * gridmapdir_otherlink, gridmapdir_urlencode, gridmapdir_newlease and
 * gridmapdir_userid are modified functions from lcmaps_gridlist.c 
 * in org.glite.security.lcmaps-plugins-basic    
 */
 
/*                                                                                                            
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or
 * http://eu-egee.org/license.html
 */

/*                                                                                                            
 * Copyright (c) 2001 EU DataGrid.                                                                             
 * For license conditions see http://www.eu-datagrid.org/license.html                                          
 *
 * Copyright (c) 2001, 2002 by 
 *     Martijn Steenbakkers <martijn@nikhef.nl>,
 *     David Groep <davidg@nikhef.nl>,
 *     NIKHEF Amsterdam, the Netherlands
 */
 

#include <utime.h>
#include <errno.h>
#include <ctype.h>
#include <dirent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <pwd.h>
#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <grp.h>
#include <malloc.h>

#include "gridmapdir.h"
#include "log.h"
#include "errors.h"

/* __________________________________________________
   local prototypes
   __________________________________________________ */

static void
gridmapdir_newlease(char *     encodedglobusidp,
                    char *     usernameprefix,
          unsigned short       matching_type,
          unsigned short *     full);
          
static char 
*gridmapdir_urlencode(char * rawstring);

static char 
*gridmapdir_otherlink(char *   firstlink);


/******************************************************************************
Function:   gridmapdir_otherlink
Description:
        find another link in MAPDIR to the same inode as firstlink
        and change the modification time of firstlink to now (so that we
        always know when this pair was last used)
        
Parameters:
        firstlink, the filename of the link we already know

Returns:
        a pointer to the other link's filename (without path) or NULL if none
        found (this is malloc'd and will need freeing)

*************************************************************************{{{ */
static char 
*gridmapdir_otherlink(char *   firstlink)
{
     int            ret;
     char           *firstlinkpath, *otherlinkdup, *otherlinkpath,
                    *gridmapdir;
     struct dirent  *gridmapdirentry;
     DIR            *gridmapdirstream;
     struct stat    statbuf;
     ino_t          firstinode;
     char           *func = "gridmapdir_otherlink()";
     
     debug_msg(8,"%s",func);
     
     gridmapdir = getenv("MAPDIR");
     if (gridmapdir == NULL) return NULL;
     
     firstlinkpath = malloc(strlen(gridmapdir) + 2 + strlen(firstlink));
     sprintf(firstlinkpath, "%s/%s", gridmapdir, firstlink);     
     ret = stat(firstlinkpath, &statbuf);
     free(firstlinkpath);   
     if (ret != 0) return NULL;
     if (statbuf.st_nlink == 1) return NULL;
     if (statbuf.st_nlink > 2)
     {
         fprintf(stderr,"gridmapdir_otherlink(): too many hardlinks found (%ld) for %s\n", (long int)statbuf.st_nlink, firstlink);
         return NULL;
     }
     
     firstinode = statbuf.st_ino; /* save for comparisons */
          
//     gridmapdirstream = opendir(gridmapdir);
     if ((gridmapdirstream = opendir(gridmapdir)) == NULL)
     {
         fprintf(stderr,"gridmapdir_otherlink(): error opening directory %s: %s\n", gridmapdir, strerror(errno));
         return NULL;
     }

     if (gridmapdirstream != NULL)
     {
         while ((gridmapdirentry = readdir(gridmapdirstream)) != NULL)
         {       
                 if (strcmp(gridmapdirentry->d_name, firstlink) == 0) continue;
           
                 otherlinkpath = malloc(strlen(gridmapdir) + 2 + 
                                        strlen(gridmapdirentry->d_name));
                 sprintf(otherlinkpath, "%s/%s", gridmapdir, 
                                            gridmapdirentry->d_name);

                 ret = stat(otherlinkpath, &statbuf);     
                 if ((ret == 0) && (statbuf.st_ino == firstinode))
                 {
                      utime(otherlinkpath, (struct utimbuf *) NULL);
                      free(otherlinkpath);
                      otherlinkdup = strdup(gridmapdirentry->d_name);
                      closedir(gridmapdirstream);     
                      return otherlinkdup;
                 }
                 else free(otherlinkpath);
         }
         
         closedir(gridmapdirstream);     
     }

     return NULL;
} /* gridmapdir_otherlink() }}} */


/* __________________________________________________
   gridmapdir_urlencode
   _______________________________________________{{{ */
static char 
*gridmapdir_urlencode(char * rawstring)
{
     int          encodedchar = 0, rawchar = 0;
     char *       encodedstring;
     /* char *func = "gridmapdir_urlencode()"; */
     
     encodedstring = (char *) malloc(3 * strlen(rawstring) + 1);
     
     if (encodedstring == NULL) return (char *) NULL;

     while (rawstring[rawchar] != '\0')
     {
           if (isalnum(rawstring[rawchar]))
           {
               encodedstring[encodedchar] = tolower(rawstring[rawchar]);
               ++rawchar;
               ++encodedchar;
           }
           else
           {
               sprintf(&encodedstring[encodedchar], "%%%02x", 
                                               rawstring[rawchar]);
               ++rawchar;
               encodedchar = encodedchar + 3;
           }        
     }

     encodedstring[encodedchar] = '\0';
     
     return encodedstring;
} /* gridmapdir_urlencode() }}} */

/* __________________________________________________
   gridmapdir_newlease
   full is set to 1 if failure is due to pool being exhausted, 
   caller must ensure it is !1 when passing
   _______________________________________________{{{ */
static void
gridmapdir_newlease(char *     encodedglobusidp,
                    char *     usernameprefix,
          unsigned short       matching_type,
          unsigned short *     full)
{
     int            ret;
     char           *userfilename, *encodedfilename, *gridmapdir;
     struct dirent  *gridmapdirentry;
     DIR            *gridmapdirstream;
     struct stat    statbuf;
     char           *func = "gridmapdir_newlease()";
     
     debug_msg(8,"%s",func);
     
     gridmapdir = getenv("MAPDIR");
     if (gridmapdir == NULL) return;

     encodedfilename = malloc(strlen(gridmapdir) + (size_t) 2 + 
                              strlen(encodedglobusidp));
     sprintf(encodedfilename, "%s/%s", gridmapdir, encodedglobusidp);
     debug_msg(8,"%s: encoded filename: %s",func,encodedfilename);
     
     ret = stat(encodedfilename, &statbuf);
     if (ret == 0)
     {
         /* encodedfilename exists, but has only one link (itself) --> remove it */
         if (statbuf.st_nlink == 1)
         {
             fprintf(stderr,"gridmapdir_newlease(): removing solitary leasename: %s\n",encodedfilename);
             unlink(encodedfilename);
         }
     }

     /*
      * Previously it was not checked if gridmapdir existed ! Do it now
      */
     if ((gridmapdirstream = opendir(gridmapdir)) == NULL)
     {
         fprintf(stderr,"gridmapdir_newlease(): error opening directory %s: %s\n", gridmapdir, strerror(errno));
         debug_msg(1,"%s: error opening directory %s: %s\n", func, gridmapdir, strerror(errno));
         return;
     }

     while ((gridmapdirentry = readdir(gridmapdirstream)) != NULL)
     {
       debug_msg(8,"%s: checking gridmapdirentry: %s",func, gridmapdirentry->d_name);
       
       /* we dont want any files that dont look like acceptable usernames */
       if ((*(gridmapdirentry->d_name) == '%') || 
           (strcmp(gridmapdirentry->d_name, "root") == 0))   continue;
       else if (*(gridmapdirentry->d_name) == '.')           continue;
       else if (index(gridmapdirentry->d_name, '~') != NULL) continue;
       else if (strncmp(gridmapdirentry->d_name, usernameprefix,
                        strlen(usernameprefix)) != 0)        continue;

       userfilename = malloc(strlen(gridmapdir) + (size_t) 2 + 
                             strlen(gridmapdirentry->d_name));
       sprintf(userfilename, "%s/%s", gridmapdir, gridmapdirentry->d_name);
       debug_msg(8,"%s: checking user: %s",func, gridmapdirentry->d_name);
       
       stat(userfilename, &statbuf);

       if (statbuf.st_nlink == 1) /* this one isnt leased yet */
       {
           ret = link(userfilename, encodedfilename);
           free(userfilename);
           if (ret != 0) 
           {
               /* link failed: this is probably because a VERY lucky
                  other process has obtained a lease for encodedfilename 
                  while we were faffing around */
               /* OR the encodedfilename is already linked to another lease name ! */
               log_msg("gridmapdir_newlease(): could not link\n");
               closedir(gridmapdirstream);
               free(encodedfilename);
               return;
           }
     
           stat(encodedfilename, &statbuf);
           if (statbuf.st_nlink > 2) 
           {
              /* two globusIDs have grabbed the same username: back off */
              unlink(encodedfilename);
              continue;
           }

           closedir(gridmapdirstream);
           free(encodedfilename);
           debug_msg(8,"%s: link worked",func);
           log_msg("new mapping\n");
           return; /* link worked ok, so return */
       }
       else 
       {
           free(userfilename); /* already in use, try next one */
       }
     }
     
     *full = 1;
     closedir(gridmapdirstream);
     free(encodedfilename);
     return; /* no unleased names left: give up */     
} /* gridmapdir_newlease() }}} */

/* __________________________________________________
   begin exposed functions
   __________________________________________________ */
     
/******************************************************************************
Function:   gridmapdir_userid
**************************************************************************{{{*/
int
gridmapdir_userid(char *     globusidp,
                  char *     usernameprefix,
                  char **    useridp,
                  char ***   unixgroups,
        unsigned short       matching_type)
{
     char *encodedglobusidp  = NULL;
     char *func = "gridmapdir_userid()";
     unsigned short full = 0;
     struct passwd *pwd;
     struct group *grp;
     
     debug_msg(8,"%s",func);
     
     if (globusidp == NULL) {
       debug_msg(1,"%s problem, globusidp is null",func);
       return GENERAL_FAILURE;
     } 
     
     if (usernameprefix == NULL) {
       debug_msg(1,"%s problem, usernameprefix is null",func);
       return GENERAL_FAILURE;
     }
    
     if (useridp == NULL) {
       debug_msg(1,"%s problem, usernameprefix is null",func);
       return GENERAL_FAILURE;
     }
     
     debug_msg(7,"%s: arg globusidp = %s",func,globusidp);
     debug_msg(7,"%s: arg usernameprefix = %s",func,usernameprefix);
     debug_msg(7,"%s: arg matching_type = %d",func,matching_type);

     if (getenv("MAPDIR") == NULL) {
       debug_msg(1,"%s problem: MAPDIR not in environment",func);
       return GENERAL_FAILURE; /* MAPDIR defined? */
     }
     if (globusidp[0] != '/') {
       debug_msg(1,"%s problem: DN does not begin with /",func);
       return BADARG; /* must be a proper subject DN */
     }
     
     encodedglobusidp = gridmapdir_urlencode(globusidp);
     debug_msg(7,"%s: encodedglobusidp = %s",func,encodedglobusidp);

     *useridp = gridmapdir_otherlink(encodedglobusidp);

     if ((*useridp != NULL) && (strncmp(usernameprefix, *useridp, strlen(usernameprefix)) != 0))
     {
         char *encodedfilename = NULL;
         char *gridmapdir      = NULL;

         /* full file path building */
         gridmapdir = getenv("MAPDIR");
         encodedfilename = malloc(strlen(gridmapdir) + (size_t) 2 + strlen(encodedglobusidp));
         sprintf(encodedfilename, "%s/%s", gridmapdir, encodedglobusidp);
 
         /* unlink every DN that comes here if the 'OVERRIDE_INCONSISTENCY' parameter is added to 'matching_type' */
         if ((matching_type&OVERRIDE_INCONSISTANCY) == OVERRIDE_INCONSISTANCY)
         {
             struct stat    statbuf;
             /* unlink the encodedglobusid (represents DN) */
             stat(encodedfilename, &statbuf);
             if (statbuf.st_nlink == 2)
             {
                 fprintf(stderr,"gridmapdir_userid(): (Re)moving hard link named \"%s\" (was linked to %s, will be linked to %s*)\n", encodedfilename, *useridp, usernameprefix);
                 unlink(encodedfilename);
             }
             else
             {
                 fprintf(stderr,"gridmapdir_userid(): Something really wrong, nr hardlinks = %ld (should be 2 for %s and lease %s)\n",
                                (long int)statbuf.st_nlink, encodedfilename, *useridp);
                 /* Cleaning */
                 if (encodedfilename)
                 {
                     free (encodedfilename);
                     encodedfilename = NULL;
                 }
                 if (*useridp)
                 {
                     free(*useridp);
                     *useridp = NULL;
                 }
                 return POOL_INDEX_FAILURE;
             }
             /* Cleaning */
             if (encodedfilename)
             {
                 free (encodedfilename);
                 encodedfilename = NULL;
             }
             if (*useridp)
             {
                 free(*useridp);
                 *useridp = NULL;
             }
         }
         else
         {
             fprintf(stderr,"gridmapdir_userid(): hard link named \"%s\" points to the wrong lease: %s\n", encodedfilename, *useridp);
             /* Cleaning */
             if (encodedfilename)
             {
                 free (encodedfilename);
                 encodedfilename = NULL;
             }
             if (*useridp)
             {
                 free(*useridp);
                 *useridp = NULL;
             }
             return POOL_INDEX_FAILURE;
         }
     }


     /* maybe no lease yet (*useridp == NULL) or the lease is inconsistent with the usernameprefix from the grid-mapfile */ 
     if (*useridp == NULL)
     {
         debug_msg(4,"%s: no lease yet or lease is inconsistent with username prefix in grid-mapfile",func);

         /* try making a lease */
         gridmapdir_newlease(encodedglobusidp, usernameprefix, matching_type, &full);

         /* check lease */ 
         *useridp = gridmapdir_otherlink(encodedglobusidp); 

         /* check if now there is a consistent lease - possibly made by someone else */
         if ((*useridp == NULL) || (strncmp(usernameprefix, *useridp, strlen(usernameprefix)) != 0)) 
         {
             debug_msg(4,"%s: still no lease, fails",func);
             free(encodedglobusidp);
             
             if (full == 1) {
               log_msg("pool full\n");
               return POOL_FULL;
             } else {
               return POOL_INDEX_FAILURE; /* still no good */
             }
         }
         
     } else {
       /* lease existed, query groups of the pool acct */
       if ((pwd = getpwnam(*useridp)) == NULL) {
         fprintf(stderr, "useridp %s not found\n", *useridp);
         return GENERAL_FAILURE;
       }
       
       if (!pwd->pw_gid) {
         debug_msg(1,"pw_gid null");
         return GENERAL_FAILURE;
       }
       
       grp = getgrgid(pwd->pw_gid);
       if (!grp) {
         fprintf(stderr, "failed inquiry to group #%d\n", pwd->pw_gid);
         log_msg("failed inquiry to group #%d\n", pwd->pw_gid);
         return GENERAL_FAILURE;
       }
       
       debug_msg(8,"%s: already leased, found primary group %s",func,grp->gr_name);
       
       
       /* secondaries: */
       /*int ngroups = getgroups (0, NULL);
       gid_t *groups = (gid_t *) malloc (ngroups * sizeof (gid_t));
       int ret = getgroups (ngroups, groups);
       if (val < 0) {
         free (groups);
       } */
       
       /* return groups in string form */
       *unixgroups = (char **) malloc (sizeof(char *) * 1);
       *unixgroups[0] = grp->gr_name;
       log_msg("already leased, found primary group %s\n",grp->gr_name);
    }

    free(encodedglobusidp);
    return SUCCESS;
} /* gridmapdir_userid() }}} */

/* __________________________________________________
   gridmapdir_unlink_dn
   
   deletes association 
   _______________________________________________{{{ */
int gridmapdir_unlink_dn(char *dn, uid_t *uid, unsigned short quarantine_flag)
{
  char *func = "gridmapdir_unlink_dn()";
  char *encodedglobusidp = NULL;
  char *leasedname = NULL;
  char *gridmapdir = NULL;
  char *quarantinepath = NULL; 
  char *leasepath = NULL;
  char *quarantinesuffix = QUARANTINE_SUFFIX;
  char *encodedglobusidpath = NULL;
  struct stat    statbuf;
  int ret;
  struct passwd *pwd;
     
  debug_msg(8,"%s",func);
     
  if (dn == NULL) {
    debug_msg(1,"%s problem, dn is null",func);
    ret = BADARG;
    goto free;
  } 
  
  if (uid == NULL) {
    debug_msg(1,"%s problem, uid * is null",func);
    ret = GENERAL_FAILURE;
    goto free;
  }
  
  if (dn[0] != '/') {
    debug_msg(1,"%s problem, dn doesn't begin with '/'",func);
    ret = BADARG; /* must be a proper subject DN */
    goto free;
  }
   
  encodedglobusidp = gridmapdir_urlencode(dn);
  if (!encodedglobusidp) {
    debug_msg(3,"%s can't encode DN: %s",func,dn);
    ret = GENERAL_FAILURE;
    goto free;
  }
  
  leasedname = gridmapdir_otherlink(encodedglobusidp);
  
  if (!leasedname) {
    debug_msg(3,"%s can't delete: no association for DN: %s",func,dn);
    fprintf(stderr,"can't delete: no association for DN: %s\n",dn);
    ret = GRIDMAPDIR_NO_ASSOCIATION;
    goto free;
  }
  
  gridmapdir = getenv("MAPDIR");
  if (gridmapdir == NULL) { 
    /* gridmapdir_otherlink() returns null if no MAPDIR, so.. */
    debug_msg(1,"%s: impossible, no MAPDIR here?",func);
    goto fail;
  }
  
  leasepath = (char *) malloc(strlen(gridmapdir) + 2 + strlen(leasedname));
  if (!leasepath) {
    debug_msg(1,"%s problem, malloc",func);
    goto fail;
  }
  sprintf(leasepath, "%s/%s", gridmapdir, leasedname);
  debug_msg(8,"%s: account = %s",func,leasepath);
  
  /* first link account file to a quarantine file */
  if (quarantine_flag & QUARANTINE) {
    debug_msg(3,"%s: attempting to quarantine before unlinking",func);
    quarantinepath = malloc(strlen(leasepath) + 2 + strlen(quarantinesuffix));
    if (!quarantinepath) {
      debug_msg(1,"%s problem, malloc",func);
      goto fail;
    }
    sprintf(quarantinepath, "%s%s", leasepath, quarantinesuffix);
    debug_msg(8,"%s: quarantine to %s",func,quarantinepath);
    
    ret = stat(quarantinepath, &statbuf);
    if (ret == 0) {
      debug_msg(1,"%s: cannot quarantine, file already exists: %s",func,quarantinepath);
      /* TODO: here, override inconsistency (touch file) or just continue */
      goto fail;
      
    }
    
    ret = link(leasepath,quarantinepath);
    if (ret) {
      debug_msg(1,"%s: cannot link %s and %s", func, leasepath, quarantinepath);
      goto fail;
    }
  }
  
  encodedglobusidpath = (char *) malloc(strlen(gridmapdir) + 2 + strlen(encodedglobusidp));
  if (!encodedglobusidpath) {
    debug_msg(1,"%s problem, malloc",func);
    goto fail;
  }
  sprintf(encodedglobusidpath, "%s/%s", gridmapdir, encodedglobusidp);
  
  debug_msg(8,"%s: encodedglobusidpath = %s",func,encodedglobusidpath);
  
  ret = unlink(encodedglobusidpath);
  if (ret) { 
    /* todo, parse errno, this is an odd inconsistency */
    debug_msg(1,"%s: cannot unlink %s", func, encodedglobusidpath);
    goto fail;
  }
  
  if ((pwd = getpwnam(leasedname)) == NULL)
  {
    debug_msg(1,"%s: very unlikely.. username %s not found?", func, leasedname);
    goto fail;
  }

  *uid = pwd->pw_uid;
  endpwent();
  
  debug_msg(8,"%s: unlinked successfully, found uid = %d", func, *uid);
  
  ret = 0;
  goto free;
  fail:
    ret = 1;
  free:
    if (leasedname) free(leasedname);
    if (encodedglobusidp) free(encodedglobusidp);
    if (leasepath) free(leasepath);
    if (quarantinepath) free(quarantinepath);
    if (encodedglobusidpath) free(encodedglobusidpath);
    return ret;
} /* gridmapdir_unlink_dn() }}} */



