/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */

/*
 * IMPORTANT: This program must be setuid root and must not be runnable by
 * any user other than the trusted globus service (or root); otherwise
 * regular users could add/delete/modify grid user accounts.  To protect
 * against security breaches due to improperly set file permissions, we
 * refuse to run if the uid is not either (a) 0 or (b) the same as the gid
 * owning the executable. (note b is disabled line 97)
 *
 * Recommended file attributes: chmod 4750 && chown root.globus
 */

#include <stdio.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/fsuid.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "config.h"
#include "log.h"
#include "util.h"
#include "errors.h"
#include "gridmapdir.h"
#include "gridmap.h"

/* prototypes */
int exec_poolindex(int argc, char *argv[]);
int exec_delete_by_dn(char *dn);
int exec_unquarantine(char *poolacct);
int privileged_execv(char *const argv[]);
int user_wipe(uid_t victim_uid);
int privileged_kill(uid_t victim_uid, int signal);
int grid_init();
int pick_pool(char *dn, char **acctprefix);
int map_group(char *globusid, char **unixgroup);
int set_primary_group(char *unixname, char *primary);

/* Global variables */
uid_t g_real_uid;
char *g_gridmapfile = NULL;
char *g_gridmapdir  = NULL;
char *g_groupmapfile = NULL;

/* expiry script path EXPIRY_SCRIPT defined in config.h */

/*
 * Advanced configurations:
 */

/* if !0, DN does not need to be in grid-mapfile, just deals with
   the gridmapdir, use ".pool" for now */
#define REQUIRE_GRIDMAP_ENTRY 1

/* if REQUIRE_GRIDMAP_ENTRY is !0, CREATE_GRIDMAP_ENTRY is relevant */

/* if !0, adds DN to configured grid-mapfile with default 
   pool mapping ".pool" for now -- not implemented */
#define CREATE_GRIDMAP_ENTRY 0

/* if !0, do not call expire_pooluser.sh, delete in C */
#define HANDLE_POOL_EXPIRE 0

/* When this setuid program handles the user pool expiry (no callout to 
   script, there are three main steps, all in poolaccts.c:
     1) kill all processes run by pool user 
     2) delete all files owned by pool user in ~ and /tmp
     3) set group membership to DEFAULT_GROUP */
     
/* Executables used when this C program wipes user. */
#define RM_PROG           "/bin/rm"
#define RMDIR_PROG        "/bin/rmdir"
#define FIND_PROG         "/usr/bin/find"
     
/* default user group to return pool account to when cleaned by C program */
#define DEFAULT_GROUP "users"


/* __________________________________________________
   MAIN
  
   Check uid and euid, temporarily drop privileges, parse command line,
   and invoke appropriate routine.
   _______________________________________________{{{ */
int main(int argc, char *argv[])
{
  char buff[16], *cp;
  struct stat statbuf;
  
  if (geteuid() != 0)
  {
    fprintf(stderr, "Must run with euid root.\n");
    return ENVIRONMENT;
  }
  
  g_real_uid = getuid();
  
  /* Don't run for arbitrary users -- only for root and globus user
     Alternatively, we could hard-code an authorized uid/username */
  if (g_real_uid != 0)
  {
    // Get the gid of the file we're running from
    snprintf(buff, 16, "/proc/%d/exe", getpid());
    if (stat(buff, &statbuf) != 0)
    {
      perror("stat");
      return ENVIRONMENT;
    }
  
  /* NOTE: on other Linux distributions uid and gid don't need to match */
    /*
    // If user's id doesn't match gid of executable, refuse to run
    if (g_real_uid != statbuf.st_gid)
    {
      fprintf(stderr, "Sorry, you are not authorized to run this program.\n");
      return ENVIRONMENT;
    }
    */
  }
  
  if (debug_init()) {
    fprintf(stderr, "cannot initialize debugging\n");
    return ENVIRONMENT;
  }
  
  if (log_init()) {
    fprintf(stderr, "cannot initialize logging\n");
    return ENVIRONMENT;
  }
  
  /* drop suid for now */
  seteuid(g_real_uid);
  
  if (grid_init()) {
    fprintf(stderr, "cannot initialize grid paths\n");
    return ENVIRONMENT;
  }
  
  if (argc >= 3 && strncmp(argv[1], "-return_poolindex",18) == 0)
  {
    return exec_poolindex(argc, argv);
  }
  else if (argc == 3 && strncmp(argv[1], "-delete_dn",11) == 0)
  {
    /* this safe_dup memory is not freed */
    return exec_delete_by_dn(safe_dup(argv[2]));
  }
  else if (argc == 3 && strncmp(argv[1], "-unquarantine",14) == 0)
  {
    /* this safe_dup memory is not freed */
    return exec_unquarantine(safe_dup(argv[2]));
  }
  else
  {
    cp = strrchr(argv[0], '/');
    fprintf(stderr,
            "Syntax: %s -return_poolindex|-delete_dn|-unquarantine [args] \n",
            cp == NULL ? argv[0] : cp + 1);
    return SYNTAX;
  }
  
  /* If exec_... succeeds, we will never reach this point */
  fprintf(stderr, "Unexpected error executing specified action.\n");
  
  return GENERAL_FAILURE;
} /* main() }}} */

/* __________________________________________________
   exec_poolindex
   _______________________________________________{{{ */

int exec_poolindex(int argc, char *argv[])
{
  char *acctprefix = NULL;
  int ret;
  char *dn = safe_dup(argv[2]); /*argc checked in main >=3*/
  char *group = NULL;
  char *unixname = (char *) malloc(SAFE_LEN);
  char *primgroup = (char *) malloc(SAFE_LEN);
  char **unixgroups = NULL;
  unsigned short matching_type = MATCH_ONLY_DN;
  struct passwd *pwd;
  struct group *grp;
  
  if (argc >= 4) {
    group = safe_dup(argv[3]);
    log_msg("exec_poolindex called with DN=%s and group=%s\n",dn,group);
  } else {
    log_msg("exec_poolindex called with DN=%s\n",dn);
  }
  
  
  /* set up variables for call to gridmapdir functions */
  if ((ret = pick_pool(dn,&acctprefix))) {
    goto exit; /* pass error up */
  }
  
  if (acctprefix == NULL) {
    fprintf(stderr, "cannot obtain pool account prefix for user\n");
    ret = ENVIRONMENT;
    goto exit;
  }
    
  /* todo: move this seteuid 'down' to more specific ops in gridmapdir.c */
  seteuid(0);
  ret = gridmapdir_userid(dn,acctprefix,&unixname, &unixgroups, matching_type);
  if (ret) {
    fprintf(stderr, "cannot obtain pool account for user\n");
    /* todo: log details */
    goto exit;
  }
  
  /* if unixgroups is set by gridmapdir_userid, it is the signal that a
     lease already existed */
  if (unixgroups) {
     /* primary */
     if (unixgroups[0]) {
       debug_msg(8,"found primary group %s",unixgroups[0]);
     }
     primgroup = unixgroups[0];
  } else {
    if (group != NULL) {
      ret = map_group(group,&primgroup); 
      if (ret) {
        fprintf(stderr, "cannot obtain group mapping\n");
        goto exit;
      }
      
      ret = set_primary_group(unixname,primgroup);
      if (ret) {
        fprintf(stderr, "cannot set group %s as primary\n", primgroup);
        goto exit;
      }
    } else {
       debug_msg(8,"no group change");
       
       if ((pwd = getpwnam(unixname)) == NULL) {
         fprintf(stderr, "unixname %s not found\n", unixname);
         ret = GENERAL_FAILURE;
         goto exit;
       }
       
       if (!pwd->pw_gid) {
         debug_msg(1,"pw_gid null");
         ret = GENERAL_FAILURE;
         goto exit;
       }
       
       grp = getgrgid(pwd->pw_gid);
       if (!grp) {
         fprintf(stderr, "failed inquiry to group #%d\n", pwd->pw_gid);
         log_msg("failed inquiry to group #%d\n", pwd->pw_gid);
         ret = GENERAL_FAILURE;
         goto exit;
       }
       
       primgroup = grp->gr_name;
       
       debug_msg(3,"found primary group %s",grp->gr_name);
       
    }
  }
   
   
  log_msg("localname=%s, group=%s for DN=%s\n",unixname,primgroup,dn);
  
  printf("%s:%s\n",unixname,primgroup);
  
  ret = SUCCESS;
  
  exit:
  if (dn != NULL) free(dn);
  if (unixname != NULL) free(unixname);
  if (primgroup != NULL) free(primgroup);
  if (group != NULL) free(group);
  if (acctprefix != NULL) free(acctprefix);
  if (unixgroups != NULL) free(unixgroups);
  
  return ret;
  
} /* exec_poolindex() }}} */

/* __________________________________________________
   exec_delete_by_dn
   _______________________________________________{{{ */
int exec_delete_by_dn(char *dn)
{
  char *func = "exec_delete_by_dn()";
  unsigned short quarantine = QUARANTINE_BASE;
  char *acctprefix = NULL;
  uid_t uid = -1;
  struct passwd *pwd;
  int ret;
  
  char *args[3];
  pid_t pid;
  int status;
  char *expiry_prog;
  struct stat statbuf;
  
  debug_msg(8,"exec_delete_by_dn()");
  
  if (dn == NULL) { 
    fprintf(stderr, "null arg to exec_delete_by_dn\n");
    return GENERAL_FAILURE;
  } else {
    log_msg("exec_delete_by_dn called with DN=%s\n",dn);
  }
  
  /* set up variables for call to gridmapdir functions */
  if ((ret = pick_pool(dn,&acctprefix))) {
    return ret; /* pass error up */
  }
  
  log_msg("Deleting by DN for %s\n",dn);
  
  /* always quarantine - if DO_QUARANTINE, don't remove 
     quarantine after delete callout */
  quarantine |= QUARANTINE;
#ifdef DO_QUARANTINE
  log_msg(" with quarantine\n");
#endif

  /* unlink DN */
  
  /* todo: move this seteuid 'down' to more specific ops in gridmapdir.c */
  seteuid(0); 
  ret = gridmapdir_unlink_dn(dn, &uid, quarantine);
  if (ret) {
    debug_msg(8,"%s: gridmapdir_unlink_dn() fails",func);
    return ret; 
  }
  debug_msg(8,"%s: uid returned: %d",func,uid);
  
  if (uid < 100) {
    debug_msg(8,"%s: not cleaning or killing for user: %d",func,uid);
    log_msg("will not clean or kill for user: %d\n",uid);
    return GENERAL_FAILURE;
  }
  
  if ((pwd = getpwuid(uid)) == NULL) {
    fprintf(stderr, "uid %d not found\n", uid);
    return GENERAL_FAILURE;
  }
  
  if (!pwd->pw_name) {
    debug_msg(1,"pw_name null");
    return GENERAL_FAILURE;
  }
  
  /* At this point the lease has expired and the account should be returned
   * to an acceptable state.  The default procedure is to callout to 
   * expire_pooluser.sh, a site specific script.  EXPIRY_PROG is defined
   * in config.h, if HANDLE_POOL_EXPIRE is defined as !0, run deletion below. */
  
  if (HANDLE_POOL_EXPIRE == 0) {
       
    expiry_prog = safe_dup(EXPIRY_PROG);
    
    /* sanity checks */
    
    /* 1. file exists and is executable? */
    ret = stat(expiry_prog, &statbuf);
    if (ret) {
      log_msg("Pool expiry program (%s) does not exist\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) does not exist\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    }
    
    if (statbuf.st_mode & S_IFREG) {
      debug_msg(8,"%s: expiry script regular file",func);
    } else {
      log_msg("Pool expiry program (%s) not a regular file\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) not a regular file\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    }
    
    if (statbuf.st_mode & S_IXUSR) {
      debug_msg(8,"%s: expiry script executable by owner",func);
    } else {
      log_msg("Pool expiry program (%s) not executable by owner\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) not executable by owner\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    }
    
    
    /*  2. file is owned by root and not modifiable by others? */
    if (statbuf.st_uid != 0) {
      log_msg("Pool expiry program (%s) not owned by root\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) not owned by root\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    }
    debug_msg(8,"%s: expiry script st_mode: %x",func,statbuf.st_mode);
    
    if (statbuf.st_mode & S_IWGRP) {
      log_msg("Pool expiry program (%s) writable by group\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) writable by group\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    } else {
      debug_msg(8,"%s: expiry script not writable by group",func);
    }
    if (statbuf.st_mode & S_IWOTH) {
      log_msg("Pool expiry program (%s) writable by world\n", expiry_prog);
      fprintf(stderr, "Pool expiry program (%s) writable by world\n", expiry_prog);
      return EXPIRY_PROG_CHECK;
    } else {
      debug_msg(8,"%s: expiry script not writable by world",func);
    }
    
       
    args[0] = expiry_prog;
    args[1] = pwd->pw_name;
    args[2] = NULL;
    
    debug_msg(3,"%s: Expiring via: %s %s",func, expiry_prog, pwd->pw_name);
   
    if ((pid = fork()) == -1)
    {
      perror("fork");
    }
    else if (pid)
    {
      while (waitpid(pid, &status, 0) == -1 && errno == EINTR);
    }
    else
    {
      /* CHANGE to call site specific script with effective id !=0 */ 
      seteuid(0); 
      setuid(0); 
      execv(args[0], args);
      perror("execv");
      return GENERAL_FAILURE;
    }
    
    /* translate return value */
    if (WIFEXITED(status)) {
      debug_msg(3,"%s: Pool expiry program exit code %d", func, WEXITSTATUS(status));
      
      if (WEXITSTATUS(status)) {
        fprintf(stderr, "Pool expiry program exit code %d.\n", WEXITSTATUS(status));
        return EXPIRY_PROG_FAILURE;
      } else {
#ifdef DO_QUARANTINE
        return SUCCESS;
#else
        return exec_unquarantine(pwd->pw_name);
#endif
      }
    } else {
      if (WIFSIGNALED(status)) {
        fprintf(stderr, "Pool expiry program died by signal %d", WTERMSIG(status));
      }         
      return EXPIRY_PROG_FAILURE;
    }
    
    
    /* end if (HANDLE_POOL_EXPIRE == 0) */
  } else {
    debug_msg(3,"%s: Expiring user %d internally",func,uid);
    
    /* 1. user's processes are killed. 
     * 2. user's files in ~ and /tmp are deleted
     * 3. account's primary group is set to the default */
    
    /* Kill any of the user's processes still running */
    if (privileged_kill(uid, SIGTERM)) {
      sleep(1);
      if (privileged_kill(uid, SIGTERM))
      {
        sleep(1);
        if (privileged_kill(uid, SIGKILL))
        {
          sleep(1);
          privileged_kill(uid, SIGKILL);
        }
      }
    }
  
    /* Wipe out the user's files */
    user_wipe(uid);
    
    /* change groups back to default */
    
    ret = set_primary_group(pwd->pw_name,DEFAULT_GROUP);
    if (ret) {
      fprintf(stderr, "cannot set group %s as primary\n", DEFAULT_GROUP);
      return ret;
    }
    debug_msg(8,"%s: set primary group of %s to %s",func, pwd->pw_name, DEFAULT_GROUP);
    
    return SUCCESS;
  } /* end if HANDLE_POOL_EXPIRE != 0 */
  
  
}  /* exec_delete_by_dn }}} */

/* __________________________________________________
   exec_unquarantine
   _______________________________________________{{{ */
   
int exec_unquarantine(char *poolacct) {
  char *func = "exec_unquarantine()";
  char *quarantinesuffix = QUARANTINE_SUFFIX;
  char *quarpath = NULL;
  struct stat statbuf;
  int ret;
  
  debug_msg(8,func);
  
  if (poolacct == NULL) { 
    fprintf(stderr, "null arg to exec_unquarantine\n");
    return GENERAL_FAILURE;
  } else {
    log_msg("exec_unquarantine called with poolacct=%s\n",poolacct);
  }
  
  quarpath = (char *) malloc(strlen(g_gridmapdir) + 3 + strlen(poolacct) + strlen(quarantinesuffix));
  if (!quarpath) {
    log_msg("%s problem, malloc",func);
    fprintf(stderr,"%s problem, malloc\n",func);
    return GENERAL_FAILURE;
  }
  sprintf(quarpath, "%s/%s%s", g_gridmapdir, poolacct,quarantinesuffix);
  
  ret = stat(quarpath, &statbuf);
  if (ret) {
    log_msg("%s: cannot unquarantine %s, file does not exist: %s",func, poolacct, quarpath);
    fprintf(stderr,"cannot unquarantine %s, file does not exist: %s", poolacct, quarpath);
    return BADARG;
  }
  
  if ((statbuf.st_nlink > 2) || (statbuf.st_nlink < 2)) {
    log_msg("%s problem, unquarantine file %s has %d links",func, quarpath, statbuf.st_nlink);
    fprintf(stderr,"%s problem, unquarantine file %s has %d links",func, quarpath, statbuf.st_nlink);
    return ENVIRONMENT;
  }
  
  seteuid(0); 
  ret = unlink(quarpath);
  if (ret) { 
    debug_msg(1,"%s: cannot unlink %s", func, quarpath);
    fprintf(stderr,"%s: cannot unlink %s\n", func, quarpath);
    return GENERAL_FAILURE;
  }
  debug_msg(8,"%s: unlinked %s",func,quarpath);
  log_msg("unlinked %s\n",quarpath);
  
  return SUCCESS;
} /* exec_unquarantine() }}} */

/* __________________________________________________
   pick_pool
   result: arg2, prefix
           arg3, pooldir
   _______________________________________________{{{ */
int pick_pool(char *dn, char **acctprefix)
{
  char *func = "pick_pool()";
  char *defaultpool = ".pool";
  char *username = NULL;
  int ret;
  DIR *pooldir;
  
  int creategridmap = CREATE_GRIDMAP_ENTRY;
  int requiregridmap = REQUIRE_GRIDMAP_ENTRY;
  
  debug_msg(8,func);
  
  /* TODO: parse lcmaps.db instead */
  
  if (requiregridmap) {
    /* parse grid-mapfile for proper prefix */ 
    if(setenv("GRIDMAP", g_gridmapfile,0) != 0) {
      debug_msg(1,"%s: setting GRIDMAP environment failed", func);
    }
    
    debug_msg(8,"%s: GRIDMAP env var: %s", func,getenv("GRIDMAP"));
    
    ret = assist_gridmap(dn, &username);
    if (username == NULL) {
      log_msg("DN not in gridmap: %s\n",dn);
      fprintf(stderr,"DN not in gridmap: %s\n",dn);
      if (creategridmap) { 
        fprintf(stderr,"creating gridmap entry not implemented\n");
        return NO_USER;
      } else { 
        return NO_USER;
      }
    }
    
    debug_msg(3,"%s: username in gridmapfile: %s",func,username);
    
    if (strncmp(username,".",1) == 0) {
      *acctprefix = safe_dup(username+1);
      debug_msg(3,"%s: using prefix: %s",func,*acctprefix);
    } else { 
      log_msg("%s: found user but no pool mapping for DN: %s\n",func,dn);
      if (creategridmap) { 
        log_msg("%s: creating gridmap entry not implemented\n",func);
        free(username);
        fprintf(stderr,"no pool for user\n");
        return NO_POOL_FOR_USER;
      } else { 
        free(username);
        fprintf(stderr,"no pool for user\n");
        return NO_POOL_FOR_USER;
      }
    }
  } else {
    *acctprefix = safe_dup(defaultpool);
    log_msg("%s: gridmap entry not required, using prefix: %s\n",func,acctprefix);
  }
  
  /* for now all pools need to be in same gridmapdir */
  if (!(pooldir = opendir(g_gridmapdir)))
  {
    perror("opendir");
    free(username);
    free(acctprefix);
    return GENERAL_FAILURE;
  }
  closedir(pooldir);
  debug_msg(5,"%s: picked %s for %s",func,g_gridmapdir,dn);
  if(setenv("MAPDIR", g_gridmapdir,1) != 0) {
    log_msg("%s: setting MAPDIR environment failed", func);
    fprintf(stderr,"%s: setting MAPDIR environment failed", func);
    free(username);
    return GENERAL_FAILURE;
  }
  
  if (username != NULL) free(username);
  return SUCCESS;
} /* pick_pool() }}} */

/* __________________________________________________
   privileged_kill
   
   If signal > 0, sends specified signal to all of a user's processes.
   Returns the number of user's processes found, or -1 upon error.
   _______________________________________________{{{ */
   
int privileged_kill(uid_t victim_uid, int signal)
{
    char *func = "privileged_kill()";
    debug_msg(8,func);
  
    DIR *procdir;
    struct dirent *de;
    struct stat statbuf;
    int pid, num_procs = 0;
    char path[17] = "/proc", *file = path + 5;
    char cmdline[26] = "/proc/", *file2 = cmdline + 6;
    char *cmdline_snip = "/cmdline";
    char cmdlinebuf[75];

    // Just a precaution... don't kill all of root's processes!!!
    if (signal && !victim_uid)
    {
        return -1;
    }

    if (!(procdir = opendir(path)))
    {
        perror("opendir");
        return -1;
    }

    *file = '/';
    ++file;

    /* We only need privileges if we may be sending signal(s) */
    if (signal > 0)
    {
        seteuid(0);
    }

    while ((de = readdir(procdir)))
    {
        if ((pid = atoi(de->d_name)))
        {
            strncpy(file, de->d_name, 10);
            file[10] = '\0';
            debug_msg(20,"%s: path: %s",func,path);
            if (stat(path, &statbuf) == 0 && statbuf.st_uid == victim_uid)
            {
                /* record cmdline */

                strncpy(file2, de->d_name, 5);
                strncpy(file2+5, cmdline_snip, 8);
                debug_msg(8,"%s: cmdline: %s",func,cmdline);
                if (stat(cmdline, &statbuf) == 0) {
                  read(open(cmdline,O_RDONLY),cmdlinebuf,75);
                  cmdlinebuf[74] = '\0';
                  log_msg("killed %s\n",cmdlinebuf);
                }
                ++num_procs;
                if (signal > 0)
                {
                    kill(pid, signal);
                }
            }
        }
    }

    if (signal > 0)
    {
        seteuid(g_real_uid);
    }

    closedir(procdir);

    return num_procs;
} /* privileged_kill() }}} */

/* __________________________________________________
   user_wipe
   
   Wipes all of the user's files and directories from 
   /tmp and home directory.
   _______________________________________________{{{ */
 
int user_wipe(uid_t victim_uid)
{
    char *func = "user_wipe()";
    debug_msg(8,func);
  
    char *args[14], grid_uid_str[10];
    pid_t pid;
    struct passwd *pwd;

    /* Get the user's home directory. */
    if ((pwd = getpwuid(victim_uid)) == NULL)
    {
        fprintf(stderr, "uid %d not found\n", victim_uid);
        return GENERAL_FAILURE;
    }

    if (!pwd->pw_dir)
    {
        return GENERAL_FAILURE;
    }

    /* Delete all files owned by user */
    sprintf(grid_uid_str, "%d", victim_uid);
    args[0] = FIND_PROG;
    args[1] = pwd->pw_dir;
    args[2] = "/tmp";
    args[3] = "-uid";
    args[4] = grid_uid_str;
    args[5] = "!";
    args[6] = "-type";
    args[7] = "d";
    args[8] = "-exec";
    args[9] = RM_PROG;
    args[10] = "-f";
    args[11] = "{}";
    args[12] = ";";
    args[13] = NULL;

    if ((pid = fork()) == -1)
    {
        perror("fork");
    }
    else if (pid)
    {
        while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
    }
    else
    {
        seteuid(0);
        setuid(victim_uid);
        execv(args[0], args);
        perror("execv");
        return GENERAL_FAILURE;
    }

    /* Remove user directories */
    args[0] = FIND_PROG;
    args[1] = pwd->pw_dir;
    args[2] = "/tmp";
    args[3] = "-depth";
    args[4] = "-uid";
    args[5] = grid_uid_str;
    args[6] = "-type";
    args[7] = "d";
    args[8] = "-exec";
    args[9] = RMDIR_PROG;
    args[10] = "{}";
    args[11] = ";";
    args[12] = NULL;

    if ((pid = fork()) == -1)
    {
        perror("fork");
    }
    else if (pid)
    {
        while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
    }
    else
    {
        seteuid(0);
        setuid(victim_uid);
        execv(args[0], args);
        perror("execv");
        return GENERAL_FAILURE;
    }

    return SUCCESS;
}  /* user_wipe() }}} */

/* __________________________________________________
   grid_init
   
   initialize base gridmapdir etc.
   _______________________________________________{{{ */
int grid_init(){ 
  debug_msg(8,"grid_init()");
  g_gridmapdir = GRIDMAPDIR;
  if (g_gridmapdir == NULL) {
    fprintf(stderr,"gridmapdir null\n");
    return ENVIRONMENT;
  }
  debug_msg(5,"gridmapdir: %s",g_gridmapdir);
  
  g_gridmapfile = GRIDMAPFILE;
  if (g_gridmapfile == NULL) {
    fprintf(stderr,"gridmapfile null\n");
    return ENVIRONMENT;
  }
  debug_msg(5,"gridmapfile: %s",g_gridmapfile);
  
  g_groupmapfile = GROUPMAPFILE;
  if (g_groupmapfile == NULL) {
    fprintf(stderr,"groupmapfile null\n");
    return ENVIRONMENT;
  }
  debug_msg(5,"groupmapfile: %s",g_groupmapfile);
  
  /* todo: check validity */
  
  return SUCCESS;
} /* grid_init() }}} */

/* __________________________________________________
   map_group
   _______________________________________________{{{ */
int map_group(char *globusid, char **unixgroup) {
  char *func = "map_group()";
  int ret;
  debug_msg(8,func);
  
  /* parse grid-mapfile for proper prefix */ 
  if(setenv("GRIDMAP", g_groupmapfile,1) != 0) {
    debug_msg(1,"%s: setting GRIDMAP environment failed", func);
  }
  debug_msg(8,"%s: GRIDMAP env var: %s", func,getenv("GRIDMAP"));
  if (globusid) {
    debug_msg(3,"%s: ID: %s", func,globusid);
  } else {
    fprintf(stderr, "no group ID");
    return BADARG;
  }
  ret = assist_gridmap(globusid, unixgroup);
  if (*unixgroup == NULL) {
    log_msg("%s: ID not in group-mapfile: %s\n",func,globusid);
    return NO_GROUP;
  }
  
  debug_msg(3,"%s: group in group-mapfile: %s",func,*unixgroup);
  
  return SUCCESS;
} /* map_group() }}} */

/* __________________________________________________
   set_primary_group
   
   TODO: check usermod's return value
   _______________________________________________{{{ */
   
int set_primary_group(char *unixname, char *primary) {
  char *func = "set_primary_group()";
  char *args[5];
  int pid;
  
  debug_msg(8,func);
  
  if ((unixname == NULL) || (primary == NULL)) { 
    return GENERAL_FAILURE;
  }
  
  args[0] = USERMOD_PROG;
  args[1] = "-g";
  args[2] = primary;
  args[3] = unixname;
  args[4] = NULL;
  
  if ((pid = fork()) == -1)
  {
    perror("fork");
  }
  else if (pid)
  {
    while (waitpid(pid, NULL, 0) == -1 && errno == EINTR);
  }
  else
  {
    seteuid(0);
    execv(args[0], args);
    perror("execv");
    return GENERAL_FAILURE;
  }
  
  return SUCCESS;
} /* set_primary_group() }}} */






