/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */
package org.globus.dynamicAccounts;

import java.util.Vector;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

import org.globus.wsrf.utils.FaultHelper;

import org.globus.dynamicAccounts.types.AccountPolicy;
import org.globus.dynamicAccounts.types.AccountFaultType;
import org.globus.dynamicAccounts.types.AccountLimitFaultType;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class DynamicAccountsUtil {

    static Log logger =
        LogFactory.getLog(DynamicAccountsUtil.class.getName());

    public static int getInteger(Integer quota) {
        if (quota != null) {
            return quota.intValue();
        }  else {
            return -1;
        }
    }

    public static Vector getPolicyVector(AccountPolicy policy) {
        if (policy == null)
            return null;
        
        Vector policyInfo = new Vector();
        String[] arrayOfPolicy = policy.getPolicy();
        if (arrayOfPolicy == null)
            return null;
        
        for (int i=0; i<arrayOfPolicy.length; i++) {
            policyInfo.add(arrayOfPolicy[i]);
        }
        return policyInfo;
    }

    public static AccountFaultType makeFault(String errMsg, Exception exp) {
        
        AccountFaultType fault = new AccountFaultType();
        FaultHelper helper = new FaultHelper(fault);
        if (errMsg != null) {
            helper.setDescription(errMsg);
        }
        if (exp != null) {
            helper.addFaultCause(exp);
        }

        return (AccountFaultType)helper.getFault();
    }


    public static AccountLimitFaultType makeLimitFault(String errMsg, Exception exp) {

        AccountLimitFaultType fault = new AccountLimitFaultType();
        FaultHelper helper = new FaultHelper(fault);
        if (errMsg != null) {
            helper.setDescription(errMsg);
        }
        if (exp != null) {
            helper.addFaultCause(exp);
        }

        return (AccountLimitFaultType)helper.getFault();
    }

    /**
     * A generalized method to run a command. Assumes all environmant 
     * variables are set so the command can be executed.
     *
     * @param command
     *        Command to run.
     * @param baseErrMesg
     *        Error message that is thrown with the exception
     * @return String 
     *         The output of the process on to standard out.
     */
    public static String runCommand(String[] command, String baseErrMesg)
        throws DynamicAccountsException {

        if (command == null) {
            logger.debug("Command cannot be null");
            throw new DynamicAccountsException("Command cannot be null");
        }

        for (int i=0; i<command.length; i++) {
            logger.debug("Command is ::: " + command[i]);
        }

        Runtime runtime = Runtime.getRuntime();
        StringBuffer outBuffer = null;
        BufferedReader out = null;
        BufferedReader err = null;
        try {
            Process process = runtime.exec(command);

            int returnCode = -1;
            try {
                returnCode = process.waitFor();
            } catch (InterruptedException exp) {
                logger.error("Interupped exp thrown ", exp);
            }
            logger.debug("Return code is " + returnCode);

            // Get process output on stdout
            out = new BufferedReader(new InputStreamReader
                (process.getInputStream()));
            if (out != null) {
                String line = out.readLine(); 
                if (line != null) {
                    outBuffer = new StringBuffer(line);
                    while (line != null) {
                        line = out.readLine();
                        if (line != null) {
                            outBuffer.append("\n" + line);
                        }
                    }
                }
            }   

            // If return code is not zero, try to get error stream also
            if (returnCode != 0) {
                err = new BufferedReader(new InputStreamReader
                    (process.getErrorStream()));
                StringBuffer errBuffer = null;
                if (err != null) {
                    String line = err.readLine(); 
                    if (line != null) {
                        errBuffer = new StringBuffer("\n" + line);
                        while (line != null) {
                            line = err.readLine();
                            if (line != null) {
                                errBuffer.append("\n" + line);
                            }
                        }
                    }
                } 
                String errString = baseErrMesg;
                if (errBuffer != null) {
                     errString = errString + errBuffer.toString();
                     logger.error("ErrStr " + errString);
                }
                if (outBuffer != null) {
                    errString = errString + outBuffer.toString();
                    logger.error("Out " + errString);
                }
                logger.error(errString);
                throw new DynamicAccountsException(command + "\n" + errString);
            }
        } catch (IOException ioe) {
            logger.error(baseErrMesg, ioe);
            throw new DynamicAccountsException(command + "\n" + baseErrMesg, 
                                               ioe);
        } finally {
            try {
                if (out != null) 
                    out.close();
                if (err != null)
                    err.close();
            } catch (IOException exp) {
                logger.debug("Could not close stream", exp);
            }
        }
        if (outBuffer != null) {
            return outBuffer.toString();
        } else {
            return null;
        }
    }

    /**
     * Specialized (though still generic) runCommand for caller to parse return
     * value on errors (commands returns nonzero).
     * @param command
     * @return String
     *         The output of the process on to standard out.
     * @throws DynamicAccountsException
     * @throws ReturnException
     */
    public static String runCommand(String[] command)
        throws DynamicAccountsException, ReturnException {

        if (command == null) {
            logger.debug("Command cannot be null");
            throw new DynamicAccountsException("Command cannot be null");
        }

        if (logger.isDebugEnabled()) {
            for (int i=0; i<command.length; i++) {
                logger.debug("Command arg " + i + " :: " + command[i]);
            }
        }

        Runtime runtime = Runtime.getRuntime();
        StringBuffer outBuffer = null;
        BufferedReader out = null;
        BufferedReader err = null;
        try {
            Process process = runtime.exec(command);

            int returnCode = -1;
            try {
                returnCode = process.waitFor();
            } catch (InterruptedException exp) {
                logger.error("Interupped exp thrown ", exp);
            }
            logger.debug("Return code is " + returnCode);

            // Get process output on stdout
            out = new BufferedReader(new InputStreamReader
                (process.getInputStream()));
            if (out != null) {
                String line = out.readLine();
                if (line != null) {
                    outBuffer = new StringBuffer(line);
                    while (line != null) {
                        line = out.readLine();
                        if (line != null) {
                            outBuffer.append("\n" + line);
                        }
                    }
                }
            }

            // If return code is not zero, try to get error stream also
            if (returnCode != 0) {
                err = new BufferedReader(new InputStreamReader
                    (process.getErrorStream()));
                StringBuffer errBuffer = null;
                if (err != null) {
                    String line = err.readLine();
                    if (line != null) {
                        errBuffer = new StringBuffer("\n" + line);
                        while (line != null) {
                            line = err.readLine();
                            if (line != null) {
                                errBuffer.append("\n" + line);
                            }
                        }
                    }
                }
                if ((errBuffer != null) && (outBuffer != null)) {
                    logger.error("ErrStr " + errBuffer.toString());
                    logger.error("OutStr " + outBuffer.toString());
                    throw new ReturnException(returnCode,
                            errBuffer.toString(), outBuffer.toString());
                } else if (errBuffer != null) {
                    logger.error("ErrStr " + errBuffer.toString());
                    throw new ReturnException(returnCode, errBuffer.toString());
                } else {
                logger.error("command returned " + returnCode + " with no stderr");
                throw new ReturnException(returnCode);
                }
            }
        } catch (IOException ioe) {
            logger.error(ioe);
            throw new DynamicAccountsException(command + "\n", ioe);
        } finally {
            try {
                if (out != null)
                    out.close();
                if (err != null)
                    err.close();
            } catch (IOException exp) {
                logger.debug("Could not close stream", exp);
            }
        }
        if (outBuffer != null) {
            return outBuffer.toString();
        } else {
            return null;
        }
    }
}
