/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */
package org.globus.dynamicAccounts.impl.authz;

import org.w3c.dom.Node;
import org.globus.wsrf.impl.security.authorization.exceptions.InvalidPolicyException;
import org.globus.wsrf.impl.security.authorization.exceptions.AuthorizationException;
import org.globus.wsrf.impl.security.authorization.exceptions.CloseException;
import org.globus.wsrf.impl.security.authorization.exceptions.InitializeException;
import org.globus.wsrf.impl.security.util.AuthUtil;
import org.globus.wsrf.security.authorization.PDPConfig;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceContextException;
import org.globus.wsrf.ResourceException;
import org.globus.wsrf.config.ContainerConfig;
import org.globus.dynamicAccounts.DynamicAccountsConstants;
import org.globus.dynamicAccounts.DynamicAccountsException;
import org.globus.dynamicAccounts.impl.service.DynamicAccountsResource;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;

import javax.security.auth.Subject;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.namespace.QName;
import java.io.File;
import java.io.IOException;
import java.io.BufferedReader;
import java.io.FileReader;
import java.util.*;

public class WorkspacePDP implements org.globus.wsrf.security.authorization.PDP {

    public static final String[] CONFIG_LOCATIONS =
        {"",".", "/etc", "/etc/grid-security"};

    private String DNconfigFileName = null;
    private File DNconfigFile = null;
    private String[] dns = null;
    private longWrapper DNconfigLastModified = new longWrapper(0);

    /* attrs authz list */
    private String attrConfigFileName = null;
    private File attrConfigFile = null;
    private String[] attrs = null;
    private longWrapper attrConfigLastModified = new longWrapper(0);

    public static final String SECURITY_CONFIG_FILE =
        "dnAuthzConfigFile";

    public static final String ATTR_SECURITY_CONFIG_FILE =
        "attrAuthzConfigFile";

    private static Log logger =
        LogFactory.getLog(WorkspacePDP.class.getName());

    public String[] getPolicyNames() {
        return null;
    }

    public Node getPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }

    public Node setPolicy(Node node)
            throws InvalidPolicyException {
        return null;
    }


    public boolean isPermitted(Subject peer,
                               MessageContext msgCtx,
                               QName op)
            throws AuthorizationException {

        boolean dnPassed = false;
        boolean attrPassed = false;

        String operation = op.toString();
        String peerIdentity = AuthUtil.getIdentity(peer);

        logger.debug("Operation " + operation + " called by subject: " + peerIdentity);


        boolean factoryMethod = false;

        /* The WorkspacePDP is configured to be used by the DynamicAccountsService
           and DynamicAccountsFactoryService only.  Instead of listing each op by
           QName, a way to check if the operation is from a call to the factory or
           not is to try and retrieve the associated resource, cast exception
           implies it is a factory method.*/
        DynamicAccountsResource resource = null;
        try {
            resource = getResource();
        } catch (ClassCastException e) {
            factoryMethod = true;
        } catch (DynamicAccountsException e) {
            throw new AuthorizationException("resource does not exist");
        }

        /* Policy now on already created accounts is self only */
        if (factoryMethod == false) {
            String owner = null;
            try {
                /* In the future, this would be resource.getAccountPolicy,
                   unless resource is switched to SecureResource */
                owner = resource.getOwnerDN();
            } catch (DynamicAccountsException e) {
                throw new AuthorizationException("problem finding associated ownerDN");
            }

            if (owner.equals(peerIdentity)) {
                logger.info("ACCEPTED: Operation " + operation + " called by subject: " + peerIdentity);
                return true;
            } else {
                logger.info("DENIED: Operation " + operation + " called by subject: " + peerIdentity);
                return false;
            }
        }

        /* If the operation is in factory op list, check DN/attrs */
        try {
            if (check(peerIdentity,
                      this.DNconfigFileName,
                      this.DNconfigFile,
                      this.dns,
                      this.DNconfigLastModified) == true) {
                dnPassed = true;
            } else {
                logger.debug("DN not in authz file: " + peerIdentity);
            }
        } catch (IOException e) {
            throw new AuthorizationException(e.toString());
        }

        CredentialInformation info =
                (CredentialInformation) msgCtx.getProperty(DynamicAccountsConstants.CREDINFO);


        Vector rolesVector = null;
        VomsCredentialInformation vomsinfo = null;

        if (info == null) {
            logger.info("cannot retrieve credential info from message context");
        } else {

            if (!(info instanceof VomsCredentialInformation)) {
                throw new AuthorizationException("credenital info from message context " +
                        "is not VOMS: incompatible PIP");
            } else {
                vomsinfo = (VomsCredentialInformation) info;
            }

            rolesVector = vomsinfo.getAttrs();

        }

        if (rolesVector == null) {
            logger.warn("cannot retrieve roles from credential information");
        } else {

            if (logger.isDebugEnabled()) {
                for (int i=0; i<rolesVector.size(); i++) {
                    logger.debug("\nRoles " + rolesVector.get(i));
                }
            }

            for (int i=0; i<rolesVector.size(); i++) {
                String attr = (String)rolesVector.get(i);
                logger.debug("checking attribute " + attr);
                try {
                    if (check(attr,
                              this.attrConfigFileName,
                              this.attrConfigFile,
                              this.attrs,
                              this.attrConfigLastModified) == false) {
                        logger.warn("secondary attribute denied: " + attr);
                    } else {
                        attrPassed = true;
                    }
                } catch (IOException e) {
                    throw new AuthorizationException(e.toString());
                }
            }
        }

        if (vomsinfo != null) {
            String VO = vomsinfo.getVO();
            logger.debug("VO " + VO);
        }

        logger.debug("DN authorized? " + dnPassed);
        logger.debug("attribute authorized? " + attrPassed);

        /* see bugzilla #2287

           if 'return false' based on attribute, the exception returned to client by
           ServiceAuthorizationChain claims DN is denied.

           But, AuthorizationException implies error in policy processing,
           so something along these lines is not a viable alternative:

           throw new AuthorizationException("primary attribute denied: " + primaryAttr); */


        if (dnPassed) {
            logger.info("ACCEPTED: Operation " + operation + " called by subject: " + peerIdentity);
            return true;
        } else if (attrPassed) {
            logger.info("ACCEPTED by attribute: Operation " + operation + " called by subject: " + peerIdentity);
            return true;
        } else {
            logger.info("DENIED: Operation " + operation + " called by subject: " + peerIdentity);
            return false;
        }
    }

    private boolean check(String peerIdentity,
                            String configFileName,
                            File configFile,
                            String[] attrs,
                            longWrapper lastModified) throws IOException {
        /* if InitializeException is ignored from initialize */
        if (configFile == null) {
            configFile = findConfigFile(configFileName);
        }
        if (configFile == null) {
            logger.warn("noSecConfig " + configFileName);
            return false;
        }
        logger.debug("Last modification time for " + configFile.getAbsolutePath()
                + ": " + configFile.lastModified());

        // If config file has been modified, re-read all user rights
        if (lastModified.value < configFile.lastModified()) {
            logger.debug(lastModified.value + " < " + configFile.lastModified());
            attrs = readConfigFile(configFile, lastModified);
        }

        for (int i = 0; i < attrs.length; i++) {
            if (attrs[i].trim().equals(peerIdentity)) {
                logger.info("ACCEPT: " + peerIdentity);
                return true;
            }
        }

        return false;
    }

    public void initialize(PDPConfig config,
                           String name,
                           String id)
            throws InitializeException {
        Object configFile = config.getProperty(name, SECURITY_CONFIG_FILE);
        if (configFile == null) {
            throw new InitializeException("no DN authz file configuration");
        }
        this.DNconfigFileName = (String) configFile;
        this.DNconfigFile = findConfigFile(this.DNconfigFileName);
        if (this.DNconfigFile != null) {
            try {
                this.dns = readConfigFile(this.DNconfigFile, this.DNconfigLastModified);
            } catch (IOException e) {
                throw new InitializeException(e.toString());
            }
        } else {
            throw new InitializeException("No DN authorization file");
        }

        configFile = config.getProperty(name, ATTR_SECURITY_CONFIG_FILE);
        if (configFile == null) {
            logger.warn("no attribute authz file configuration");
            return;
        }
        this.attrConfigFileName = (String) configFile;
        this.attrConfigFile = findConfigFile(this.attrConfigFileName);
        if (this.attrConfigFile != null) {
            try {
                this.attrs = readConfigFile(this.attrConfigFile, this.attrConfigLastModified);
            } catch (IOException e) {
                throw new InitializeException(e.toString());
            }
        } else {
            throw new InitializeException("No attribute authorization file");
        }
    }

    private String[] readConfigFile(File authzFile, longWrapper authzModified) throws IOException {
        TextFile file = new TextFile(authzFile.getAbsolutePath());
        authzModified.value = authzFile.lastModified();

        String[] result = new String[file.size()];

        for (int i = 0; i < file.size(); i++) {

            result[i] = (String) file.get(i);
        }
        if (logger.isDebugEnabled()) {
            for (int i = 0; i < result.length; i++) {
                logger.debug("result[" + i + "] = " + result[i]);
            }
        }

        return result;
    }

    public void close() throws CloseException {
    }



    private File findConfigFile(String fileName) {
        File configFile = null;
        for(int i = 0; i < CONFIG_LOCATIONS.length; i++) {
            String filePath = (CONFIG_LOCATIONS[i].equals("") ? fileName :
                               CONFIG_LOCATIONS[i] + File.separator +
                               fileName);
            configFile = new File(filePath);
            logger.debug("Trying authz file: " + configFile.getAbsolutePath());
            if (configFile.exists()) {
                logger.debug("Security config file found: " +
                             configFile.getAbsolutePath());
                break;
            } else {
                configFile = null;
            }
        }

        /* As a last resort, check globus location */
        if (configFile == null) {
            String globusLocation = ContainerConfig.getGlobusLocation();
            String filePath = globusLocation + File.separator + fileName;
            configFile = new File(filePath);
            logger.debug("Trying authz file: " + configFile.getAbsolutePath());
            if (configFile.exists()) {
                logger.debug("Security config file found: " +
                             configFile.getAbsolutePath());
            } else {
                configFile = null;
            }
        }

        return configFile;
    }

    /* used because you can't change value of a Long object and therefore
       cannot pass by reference */
    private class longWrapper {
        long value;

        longWrapper(long val) {
            this.value = val;
        }
    }

    private class TextFile extends ArrayList {
        //via Eckel
        private TextFile(String fileName) throws IOException {
            super(Arrays.asList(gRead(fileName).split("\n")));
        }
    }

    private static String gRead(String fileName) throws IOException {
        StringBuffer sb = new StringBuffer();
        BufferedReader in =
        new BufferedReader(new FileReader(fileName));
        String s;
        while((s = in.readLine()) != null) {
            sb.append(s);
            sb.append("\n");
        }
        in.close();
        return sb.toString();
    }


    private DynamicAccountsResource getResource()
        throws DynamicAccountsException {
        try {
            ResourceContext context = ResourceContext.getResourceContext();
            return (DynamicAccountsResource)context.getResource();
        } catch (ResourceContextException exp) {
            String err = "Error getting resource";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        } catch (ResourceException exp) {
            String err = "Error getting factory resource";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        }
    }

}
