/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */
package org.globus.dynamicAccounts.impl.factory;

import org.globus.wsrf.impl.SimpleResourceKey;

import org.globus.dynamicAccounts.types.AccountInformation;

import org.globus.dynamicAccounts.DynamicAccountsException;
import org.globus.dynamicAccounts.DynamicAccountsConstants;
import org.globus.dynamicAccounts.LimitException;

import org.globus.dynamicAccounts.impl.service.DynamicAccountsHome;


import org.globus.wsrf.Constants;
import org.globus.wsrf.ResourceKey;

import javax.naming.InitialContext;

import javax.xml.namespace.QName;

import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;

import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.SimpleResourcePropertySet;

import org.globus.wsrf.security.SecurityManager;


import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.util.Properties;
import java.util.Enumeration;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

// Store last modified timestamp of properties file and reload and stuff.
public class DynamicAccountsFactoryResourceImpl 
    implements DynamicAccountsFactoryResource {

    static Log logger = 
        LogFactory.getLog(DynamicAccountsFactoryResourceImpl.class.getName());
    
    private ResourcePropertySet propSet;
    private SimpleResourceKey key = null;
    private Integer defaultTtl;
    private Properties properties;
    private String authzFile = null;

    public synchronized void initialize(SimpleResourceKey key, Integer ttl,
                                        String host, String authzInfoFile)
        throws Exception {
        this.key = key;
        this.defaultTtl = ttl;
        this.authzFile = authzInfoFile;
        if (this.authzFile == null) {
            this.authzFile = (String)key.getValue() + AUTHZ_FILE_SUFFIX;
        }
        logger.debug("Resource created for mechanism " + key.getValue());
        this.propSet = new SimpleResourcePropertySet(RP_SET);
        ResourceProperty prop = 
            new SimpleResourceProperty(new QName(DynamicAccountsConstants.NS, 
                                                 "DefaultTtl"));
        prop.add(this.defaultTtl);
        this.propSet.add(prop);
        prop = 
            new SimpleResourceProperty(new QName(DynamicAccountsConstants.NS, 
                                                 "Host"));
        prop.add(host);
        this.propSet.add(prop);

        properties = new Properties();
        File propFile = new File(this.authzFile);
        if (propFile.exists()) {
            properties.load(new FileInputStream(propFile));
        }
    }

    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    // Create account resource using the resource home in DAS. The
    // creation mechanism used is passed as a parameter to the create method.
    public ResourceKey createAccount(AccountInformation accountInfo, String callerUserName)
        throws DynamicAccountsException, LimitException {
        
        DynamicAccountsHome home = null;
        try {
            home = getServiceHome();
        } catch (Exception exp) {
            throw new DynamicAccountsException("Error getting service home", 
                                               exp);
        }

        String callerDN = SecurityManager.getManager().getCaller();
        logger.debug("caller is " + callerDN);
        // type casting to string here. Key type should not be altered
        // in jndi config.
        ResourceKey serviceResourceKey = 
            home.create(accountInfo, (String)key.getValue(), callerDN, callerUserName, defaultTtl.intValue());
        String localUserName = (String)serviceResourceKey.getValue();
        if (accountInfo.getPolicy() != null) {
            String[] policy = accountInfo.getPolicy().getPolicy();
            if (policy != null) {
                synchronized(properties) {
                    for (int i=0; i<policy.length; i++) {
                        // Check if the dn exist in properties
                        String value = properties.getProperty(policy[i]);
                        if (value != null) {
                            // append local user name
                            properties.setProperty(policy[i], value + ","
                                                   + localUserName);
                        } else {
                            // add new entry
                        properties.setProperty(policy[i], localUserName);
                        }
                        writePropertiesToFile();
                    }
                }
            }
        }
        return serviceResourceKey;
    }

    public void addAuthz(String dn, String localName)
        throws DynamicAccountsException {

        String value = (String)this.properties.get(dn);
        // if value already exists, return
        if ((value != null) && (value.indexOf(localName) != -1)) {
            return;
        }

        if (value != null) {
            localName = value + "," + localName;
        }
        logger.debug("new value is " + localName);
        synchronized(properties) {
            this.properties.setProperty(dn, localName);
        }
        writePropertiesToFile();
    }

    public void removeAuthz(String dn, String localName)
        throws DynamicAccountsException {
        removeAuthzEntry(dn, localName);
        writePropertiesToFile();
    }

    public void removeAuthzEntry(String dn, String localName)
        throws DynamicAccountsException {

        logger.debug("DN " + dn + " LocalName " + localName);
        String value = (String)this.properties.get(dn);
        logger.debug("Old value " + value);
        if ((value == null) || (value.indexOf(localName) == -1)) {
            return;
        }
        synchronized(properties) {
            boolean allgone = false;

            // deletes repeats of localName
            while (allgone == false) {
                if (value.equals(localName)) {
                    this.properties.remove(dn);
                    allgone = true;
                } else{
                    if (value.indexOf(localName) == -1) {
                        allgone = true;
                    } else {
                        int startIndex = value.indexOf(localName);
                        int endIndex = value.indexOf(",", startIndex);
                        if (endIndex == -1)
                            endIndex = value.length();
                        String newValue = null;
                        if (startIndex == 0) {
                            logger.debug("startIndex is zero");
                            newValue = value.substring(endIndex + 1, value.length());
                        } else if (endIndex == value.length()) {
                            logger.debug("end index is length");
                            newValue = value.substring(0, startIndex-1);
                        } else {
                            newValue = value.substring(0, startIndex) +
                                value.substring(endIndex + 1, value.length());
                        }
                        logger.error("New values" + dn + " = " + newValue);
                        this.properties.setProperty(dn, newValue);
                        value = newValue;
                    }
                }
            }
        }
    }

    public void deleteAccountAuthz(String localName)
        throws DynamicAccountsException {

        logger.debug("localname " + localName);
        for (Enumeration e=this.properties.keys(); e.hasMoreElements();) {
            String dn = (String)e.nextElement();
            removeAuthzEntry(dn, localName);
        }
        writePropertiesToFile();
    }

    protected DynamicAccountsHome getServiceHome() 
        throws DynamicAccountsException{
        
        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         DynamicAccountsConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (DynamicAccountsHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        DynamicAccountsConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new DynamicAccountsException(exp);
        }
    }

    private void writePropertiesToFile() throws DynamicAccountsException {

        try {
            FileOutputStream os =
                new FileOutputStream(this.authzFile);
            properties.store(os,
                             "Mapping DN=localUserAccount. "
                             + "DONOT MODIFY FILE BY HAND");
        } catch (IOException exp) {
            String errStr = "Error writing authz info to file";
            logger.error(errStr);
            throw new DynamicAccountsException(errStr);
        }
    }
}
