/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */
package org.globus.dynamicAccounts.impl.factory;

import org.apache.axis.message.addressing.EndpointReferenceType;

import org.globus.dynamicAccounts.DynamicAccountsUtil;
import org.globus.dynamicAccounts.DynamicAccountsException;
import org.globus.dynamicAccounts.DynamicAccountsConstants;
import org.globus.dynamicAccounts.LimitException;

import org.globus.dynamicAccounts.types.*;

import org.globus.wsrf.container.ServiceHost;

import org.globus.wsrf.ResourceKey;
import org.globus.wsrf.ResourceContext;
import org.globus.wsrf.ResourceException;

import org.globus.wsrf.ResourceContextException;

import org.globus.wsrf.utils.AddressingUtils;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.globus.gsi.jaas.UserNamePrincipal;

import java.util.Set;
import java.util.Iterator;

import javax.security.auth.Subject;

import org.apache.axis.MessageContext;

import org.globus.wsrf.impl.security.authentication.Constants;

public class DynamicAccountsFactoryService {

    static Log logger = 
        LogFactory.getLog(DynamicAccountsFactoryService.class.getName());

    // Invoke initialize on factory home to set up resources for each
    // creation mechanism.
    public DynamicAccountsFactoryService() throws Exception {
        
        ResourceContext ctx = ResourceContext.getResourceContext();
        DynamicAccountsFactoryHome home = 
            (DynamicAccountsFactoryHome)ctx.getResourceHome();
        home.initialize();
    }

    private String getUserName() {
	// user name of requester, as mapped in the configured grid map file.
	MessageContext msgCtx  = MessageContext.getCurrentContext();
	Subject subject = (Subject)msgCtx.getProperty(Constants.PEER_SUBJECT);
	Set principalSet = subject.getPrincipals();
	String userName = null;
	if (principalSet != null) {
	    Iterator it = principalSet.iterator();
	    while (it.hasNext()) {
		Object principal = it.next();
		if (principal instanceof UserNamePrincipal) {
		    userName = ((UserNamePrincipal)principal).getName();
		}
	    }
	}  
	logger.debug("User name is " + userName); 
	return userName;
    }

    // Invoke createResource on appropriate factory resource.
    public EndpointReferenceType createAccount(AccountInformation accountInfo)
        throws AccountFaultType, AccountLimitFaultType {
        
        logger.debug("Create account " + accountInfo.getClass().getName());
        DynamicAccountsFactoryResource resource = null;
        try {
            resource = getFactoryResource();
        } catch (DynamicAccountsException exp) {
            throw DynamicAccountsUtil.makeFault("Error getting resource", exp);
        }

        // Will not be null, since authz will fail if its null;
        String userName = getUserName();

        ResourceKey key = null;
        try {
            key = resource.createAccount(accountInfo, userName);
        } catch (Exception exp) {
            logger.error("Error creating resource", exp);

            if (exp instanceof LimitException) {
                throw DynamicAccountsUtil.makeLimitFault("account limit reached",exp);
            } else {
                throw DynamicAccountsUtil.makeFault("error", exp);
            }
        }

        EndpointReferenceType epr = getEPR(key);

        return epr;
    }

    // Invoke createResource on appropriate factory resource.
    public AccountSetEPR createAccountSet(CreateAccountSet accountsInfo)
        throws AccountFaultType, AccountLimitFaultType {

        DynamicAccountsFactoryResource resource = null;
        try{
            resource = getFactoryResource();
        } catch (DynamicAccountsException exp) {
            throw DynamicAccountsUtil.makeFault("Error getting resource", exp);
        }
	    // Will not be null, since authz will fail if its null;
	    String userName = getUserName();

        AccountInformation accounts[] =
            accountsInfo.getAccounts();
        EndpointReferenceType EPRs[] =
            new EndpointReferenceType[accounts.length];
        logger.debug("numaccts " + accounts.length);
        for (int i=0; i < accounts.length; i++) {
            try {
                EPRs[i] =
                    getEPR(resource.createAccount(accounts[i], userName));
                logger.debug("EPR " + EPRs[i]);
            } catch (LimitException exp) {
                 String err = "Error creating multiple account, one of the " +
                         "accounts is a duplicate";
                //TODO: decide on behavior here, continue trying to create other accounts?

                throw DynamicAccountsUtil.makeLimitFault(err,exp);

            } catch (DynamicAccountsException exp) {
                String err = "Error creating multiple account";
                logger.error(err);
                throw DynamicAccountsUtil.makeFault(err, exp);
            }
        }
        return new AccountSetEPR(EPRs);
    }

    private DynamicAccountsFactoryResource getFactoryResource() 
        throws DynamicAccountsException {
        try {
            ResourceContext context = ResourceContext.getResourceContext();
            return (DynamicAccountsFactoryResource)context.getResource();
        } catch (ResourceContextException exp) {
            String err = "Error getting factory resource";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        } catch (ResourceException exp) {
            String err = "Error getting factory resource";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        }
    }

    protected EndpointReferenceType getEPR(ResourceKey key) 
        throws AccountFaultType {

        EndpointReferenceType epr = null;
        try {
            String addr = ServiceHost.getBaseURL() 
                + DynamicAccountsConstants.SERVICE_PATH;
            epr = AddressingUtils.createEndpointReference(addr, key);
        } catch (Exception exp) {
            String err = "Error creating endpoint reference";
            logger.error(err, exp);
            throw DynamicAccountsUtil.makeFault(err, exp);
        }
        return epr;
    }
}
