/*
 * This file or a portion of this file is licensed under the terms of the
 * Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without modifications,
 * you must include this notice in the file.
 */
package org.globus.dynamicAccounts.impl.service;

import org.globus.dynamicAccounts.DynamicAccountsException;
import org.globus.dynamicAccounts.DynamicAccountsConstants;
import org.globus.dynamicAccounts.DynamicAccountsUtil;
import org.globus.dynamicAccounts.impl.factory.DynamicAccountsFactoryHome;
import org.globus.dynamicAccounts.impl.factory.DynamicAccountsFactoryResource;
import org.globus.dynamicAccounts.impl.authz.CredentialInformation;
import org.globus.dynamicAccounts.impl.authz.VomsCredentialInformation;
import org.globus.dynamicAccounts.types.AccountInformation;
import org.globus.dynamicAccounts.types.VOMSAttributes;

import org.globus.wsrf.*;
import org.globus.wsrf.impl.SimpleResourcePropertySet;
import org.globus.wsrf.impl.SimpleResourceProperty;
import org.globus.wsrf.impl.ReflectionResourceProperty;
import org.globus.wsrf.impl.SimpleResourceKey;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.axis.MessageContext;

import javax.xml.namespace.QName;
import javax.naming.InitialContext;
import java.util.Calendar;
import java.util.Vector;

public class DynamicAccountsResourceImpl implements DynamicAccountsResource {

    /* immutable, passed to initialize() */
    protected String creationMechanism;
    protected String callerDN;
    protected String callerUsername; //used by unix resource

    /* set in initialize() */
    protected QName RP_SET = new QName(DynamicAccountsConstants.NS, "DASRPSet");
    protected ResourcePropertySet propSet;
    protected ResourceProperty prop = null;
    protected int factoryTtl = -1; //the default ttl from factory

    /* initialized in createAccount() */
    int ttl = -1;
    protected Calendar terminationTime = null;
    protected Vector policy = null;
    protected String [] secondaryGroups = null;

    /* immutable, set in createAccount() */
    protected String localName = null;
    protected String primaryGroup = null;
    protected Vector rolesVector = null;
    protected String [] vomsAttributes = null;
    protected String VO = null;
    protected String primaryAttr = null; //this field isn't necessary anymore, but convenient

    /* immutable, set in handleCallerCredential() */
    protected boolean usingVoms = false;

    static Log logger =
        LogFactory.getLog(DynamicAccountsResourceImpl.class.getName());

    /*
     *  Begin public methods
     */

    public void initialize(String _creationMechanism,
                           String _callerDN,
                           String _callerUserName,
                           String globusLocation,
                           int defaultTTl)
            throws DynamicAccountsException {

        if (_creationMechanism == null) {
            throw new DynamicAccountsException("creation mechanism cannot be null");
        }
        if (_callerDN == null) {
            throw new DynamicAccountsException("caller DN cannot be null");
        }
        if (_callerUserName == null) {
            logger.debug("_callerUserName = " + _callerUserName);
        }
        if (globusLocation == null) {
            throw new DynamicAccountsException("globusLocation cannot be null");
        }

        this.creationMechanism = _creationMechanism;
        this.callerDN = _callerDN;
        this.callerUsername = _callerUserName;
        this.factoryTtl = defaultTTl;

        globusLocation = globusLocation.trim();

        this.propSet = new SimpleResourcePropertySet(RP_SET);
        QName creationQname = new QName(DynamicAccountsConstants.NS,
                                        "CreationMechanism");
        prop = new SimpleResourceProperty(creationQname);
        prop.add(this.creationMechanism);
        this.propSet.add(prop);
    }

    public String createAccount(AccountInformation accountInfo)
            throws DynamicAccountsException {
        // TTL is assumed to be in minutes
        if (accountInfo.getTtl() != null) {
            this.ttl = accountInfo.getTtl().intValue();
        } else {
            this.ttl = this.factoryTtl;
        }
        if (this.ttl < 1) {
            throw new DynamicAccountsException
                        ("invalid time to live, < 1 minute");
        }
        this.terminationTime = Calendar.getInstance();
        this.terminationTime.add(Calendar.MINUTE, this.ttl);

        // This needs to be part of gridmap, convert to string list
        this.policy =
            DynamicAccountsUtil.getPolicyVector(accountInfo.getPolicy());

        MessageContext msgCtx = MessageContext.getCurrentContext();
        CredentialInformation info =
                (CredentialInformation)msgCtx.getProperty(DynamicAccountsConstants.CREDINFO);
        if (info == null) {
            logger.debug("no credential information in msgCtx");
        } else {
            handleCallerCredential(info);
        }

        createAccountUsingMech();

        setResourceProperties();

        logger.info("WORKSPACE CREATED: Local name " + this.localName + ", ttl " + this.ttl + ", ownerDN " + this.callerDN);
        return this.localName;
    }

    private void handleCallerCredential(CredentialInformation info)
            throws DynamicAccountsException {

        if (info instanceof VomsCredentialInformation) {
            logger.debug("using voms cred");

            VomsCredentialInformation vomsinfo =
                    (VomsCredentialInformation) info;

            this.rolesVector = vomsinfo.getAttrs();

            //note: if there are no roles (e.g., when a regular proxy is used),
            //      the vector object will be present, but have size 0
            if (this.rolesVector == null) {
                throw new DynamicAccountsException
                        ("cannot retrieve roles from credential information");
            }

            if (rolesVector.size() > 0) {
                this.vomsAttributes = new String[rolesVector.size()];
                for (int i=0; i<rolesVector.size(); i++) {
                    String role = (String) rolesVector.get(i);
                    logger.debug("\nRole " + role);
                    this.vomsAttributes[i] = role;
                }
                this.primaryAttr = (String) this.rolesVector.get(0);
            }

            //ignore vomsinfo.getHostport()

            this.VO = vomsinfo.getVO();

            this.usingVoms = true;
        }

    }


    // Change the file with authz info to add this dn.
    public void addAuthzIdentity(String str) throws DynamicAccountsException {
        if (this.policy == null)
            this.policy = new Vector();
        this.policy.add(str);
        DynamicAccountsFactoryResource factoryResource = getFactoryResource();
        if (factoryResource == null) {
            String err = "Factory resource for " + this.creationMechanism +
                " not found";
            logger.error(err);
            throw new DynamicAccountsException(err);
        }
        factoryResource.addAuthz(str, this.localName);

        if (this.policy != null) {
            QName qname = new QName(DynamicAccountsConstants.NS, "Policy");
            prop = new SimpleResourceProperty(qname);
            prop.add(this.policy);
            this.propSet.add(prop);
        }
    }

    // Reflect this change in actual unix account
    public void removeAuthzIdentity(String str) throws DynamicAccountsException {
        if (this.policy != null) {
            this.policy.remove(str);
        }
        DynamicAccountsFactoryResource factoryResource = getFactoryResource();
        if (factoryResource == null) {
            String err = "Factory resource for " + this.creationMechanism +
                " not found";
            logger.error(err);
            throw new DynamicAccountsException(err);
        }
        factoryResource.removeAuthz(str, this.localName);
        if (this.policy != null) {
            QName qname = new QName(DynamicAccountsConstants.NS, "Policy");
            prop = new SimpleResourceProperty(qname);
            prop.add(this.policy);
            this.propSet.add(prop);
        }
    }

    public String getLocalAccountName() throws DynamicAccountsException {
        return this.localName;
    }

    public String getOwnerDN() throws DynamicAccountsException {
        return this.callerDN;
    }

    public VOMSAttributes getVOMSAttributes() throws DynamicAccountsException {
        VOMSAttributes attrs = new VOMSAttributes();
        attrs.setFqan(this.vomsAttributes);
        return attrs;
    }

    // Resource properties
    public ResourcePropertySet getResourcePropertySet() {
        return this.propSet;
    }

    // Resource lifetime interface
    public void setTerminationTime(Calendar _terminationTime) {
        this.terminationTime   = _terminationTime;
    }

    public Calendar getTerminationTime() {
        return this.terminationTime;
    }

    public Calendar getCurrentTime() {
        return Calendar.getInstance();
    }

    public void deleteAccount() throws DynamicAccountsException {
        ResourceKey key = DynamicAccountsHome.getKey(this.localName);
        logger.debug("deleteAccount called with key = " + key.getName().toString());
        DynamicAccountsHome home = getServiceHome();
        try {
	    home.remove(key);
        } catch (ResourceException exp) {
            String err = "Deleting account resource failed";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        }
    }

    public void remove() throws ResourceException {
        try {
            // delete entries from authz info file
            DynamicAccountsFactoryResource factoryResource =
                getFactoryResource();
            if (factoryResource == null) {
                String err = "Factory resource for " + this.creationMechanism
                    + " not found";
                logger.error(err);
                throw new ResourceException(err);
            }
            logger.debug("Calling deleteAccount ");
            factoryResource.deleteAccountAuthz(this.localName);
        } catch (DynamicAccountsException exp) {
            logger.error(exp);
            throw new ResourceException(exp);
        }


        /* Delete account on system -- errors result in resource removal
           as of tech_preview_1.  Before, a system account removal failure
           (resulting in a quarantined pool account) did not mean the
           resource should be deleted (only the policy associated with the
           account). */
        try {
            // delete account
            deleteAccountUsingMech();
        } catch (DynamicAccountsException exp) {
            logger.info("Problem removing system account -- resource destroyed");
        }

        logger.info("WORKSPACE REMOVED: Local name " + this.localName +  ", ownerDN " + this.callerDN);

    }


    /*
    *  END public methods
    */


    /*
     * BEGIN setuid interface related methods
     */


    /* note: constructGroupMapInput() is no longer necessary, LCMAPS handles
       the FQAN mappings */

    protected void createAccountUsingMech() throws DynamicAccountsException {
        logger.warn("create account no-op");
    }

    protected void deleteAccountUsingMech() throws DynamicAccountsException {
        logger.warn("remove account no-op");
    }

    /*
     * END setuid interface related methods
     */

    /*
     * BEGIN Framework related methods
     */

    protected void setResourceProperties() throws DynamicAccountsException {
        QName qname = new QName(DynamicAccountsConstants.NS, "AccountType");
        prop = new SimpleResourceProperty(qname);
        prop.add(DynamicAccountsConstants.IND_ACCOUNT);
        this.propSet.add(prop);

        qname = new QName(DynamicAccountsConstants.NS, "LocalName");
        prop = new SimpleResourceProperty(qname);
        prop.add(this.localName);
        this.propSet.add(prop);

        qname = new QName(DynamicAccountsConstants.NS, "TTL");
        prop = new SimpleResourceProperty(qname);
        prop.add(new Integer(this.ttl));
        this.propSet.add(prop);

        if (this.policy != null) {
            qname = new QName(DynamicAccountsConstants.NS, "Policy");
            prop = new SimpleResourceProperty(qname);
            prop.add(this.policy);
            this.propSet.add(prop);
        }

        if (this.vomsAttributes != null) {
            qname = new QName(DynamicAccountsConstants.NS, "VOMSAttributes");
            prop = new SimpleResourceProperty(qname);
            prop.add(this.vomsAttributes);
            this.propSet.add(prop);
        }

        try {
            prop =
                new ReflectionResourceProperty(WSRFConstants.TERMINATION_TIME,
                                               this);
            this.propSet.add(prop);
            prop = new ReflectionResourceProperty(WSRFConstants.CURRENT_TIME,
                                                  this);
            this.propSet.add(prop);
        } catch (Exception exp) {
            String err = "Error setting resource properties";
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        }
    }



    protected DynamicAccountsFactoryResource getFactoryResource()
        throws DynamicAccountsException {

        // get key to factory resource for this creation mechanism
        SimpleResourceKey factoryResourceKey =
            DynamicAccountsFactoryHome.getResourceKey(this.creationMechanism);
        DynamicAccountsFactoryHome factoryHome = getFactoryHome();
        DynamicAccountsFactoryResource factoryResource = null;
        try {
            factoryResource = (DynamicAccountsFactoryResource)factoryHome
                .find(factoryResourceKey);
        } catch (ResourceException exp) {
            String err = "Error getting factory resource foor "
                + this.creationMechanism;
            logger.error(err, exp);
            throw new DynamicAccountsException(err, exp);
        }
        return factoryResource;
    }



    protected DynamicAccountsFactoryHome getFactoryHome()
        throws DynamicAccountsException {

        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         DynamicAccountsConstants.FACTORY_PATH +
                         Constants.HOME_NAME);
            return (DynamicAccountsFactoryHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        DynamicAccountsConstants.FACTORY_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new DynamicAccountsException(exp);
        }
    }

    protected DynamicAccountsHome getServiceHome()
        throws DynamicAccountsException {

        try {
            InitialContext context = new InitialContext();
            logger.debug(Constants.JNDI_SERVICES_BASE_NAME +
                         DynamicAccountsConstants.SERVICE_PATH +
                         Constants.HOME_NAME);
            return (DynamicAccountsHome)context
                .lookup(Constants.JNDI_SERVICES_BASE_NAME +
                        DynamicAccountsConstants.SERVICE_PATH +
                        Constants.HOME_NAME);
        } catch (Exception exp) {
            throw new DynamicAccountsException(exp);
        }
    }

    /*
     * END Framework related methods
     */


}
