package org.globus.dynamicAccounts.impl.service;

import org.globus.dynamicAccounts.DynamicAccountsException;
import org.globus.dynamicAccounts.DynamicAccountsUtil;
import org.globus.dynamicAccounts.ReturnException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.File;
import java.util.StringTokenizer;
import java.util.ArrayList;

public class LCMAPSResource extends DynamicAccountsResourceImpl {

    static Log logger =
        LogFactory.getLog(LCMAPSResource.class.getName());

    /* set in initialize() */
    protected String executable = null;

    protected final static String BASE_COMMAND = "bin" + File.separator
        + "lcmaps_poolindex";
    protected final static String POOLINDEX_COMMAND = "-return_poolindex";
    protected final static String DELETE_BY_POOLINDEX_COMMAND = "-unlease_poolindex";

    /* set in createAccountUsingMech(), used to make life easier with interface
       to LCMAPS.  Also used as resource key.  Meant to be immutable once set.  */
    protected String poolindex;


    //overriding to set executable
    public void initialize(String _creationMechanism,
                       String _callerDN,
                       String _callerUserName,
                       String globusLocation,
                       int defaultTtl)
        throws DynamicAccountsException {

        super.initialize(_creationMechanism, _callerDN, _callerUserName, globusLocation, defaultTtl);

        if (globusLocation.endsWith(File.separator)) {
            executable = globusLocation + BASE_COMMAND;
        } else {
            executable = globusLocation + File.separator + BASE_COMMAND;
        }
        logger.debug("Executable " + executable);

    }

    /*
     * BEGIN setuid interface related methods
     */

    protected void createAccountUsingMech() throws DynamicAccountsException {
        String command[] = null;

        /* no primaryAttr means no attributes at all -- either VOMS is not being
           employed or proxy has no attributes */
        if (this.primaryAttr == null) {
            command =  new String[3];
            command[0] = executable;
            command[1] = POOLINDEX_COMMAND;
            command[2] = this.callerDN;
        } else if ((this.rolesVector != null) && (this.rolesVector.size() > 0)) {
            command =  new String[3 + this.rolesVector.size()];
            command[0] = executable;
            command[1] = POOLINDEX_COMMAND;
            command[2] = this.callerDN;
            for (int i = 0 ; i < this.rolesVector.size(); i++) {
                String test = (String) this.rolesVector.get(i);
                if (test != null) {
                    command[i+3] = test;
                } else {
                    throw new DynamicAccountsException("found a null attribute " +
                            "in rolesVector: index #" + i);
                }
            }
        } else {
            // this can only happen if VomsCredentialPIP is used but does not
            // parse roles for some reason *and* primaryAttr is set somehow
            // (a developer extending is the only way this could happen)
            throw new DynamicAccountsException("scheme not supported");
        }

        String returnString = null;
        try {
            returnString = DynamicAccountsUtil.runCommand(command);
        } catch (ReturnException e) {
            translateReturnValue(e);
        }


        logger.debug("Return string:\n" + returnString);

        // Required: poolindex, unixname, primary unix group
        // Optional: secondary groups
        StringTokenizer st = new StringTokenizer(returnString,"\n");
        if (st.hasMoreTokens()) {
            this.poolindex = st.nextToken();
            if (st.hasMoreTokens()) {
                this.localName = st.nextToken();
                if (st.hasMoreTokens()) {
                    this.primaryGroup = st.nextToken();
                } else {
                    throw new DynamicAccountsException("no primary group returned from pool impl");
                }
                ArrayList secs = new ArrayList();
                while (st.hasMoreTokens()) {
                    secs.add(st.nextToken());
                }
                if (secs.size() > 0) {
                    String[] tmp = new String[secs.size()];
                    this.secondaryGroups = (String[]) secs.toArray(tmp);
                }
            } else {
                throw new DynamicAccountsException("no poolindex returned from pool impl");
            }
        } else {
            throw new DynamicAccountsException("no information returned from pool " +
                    "impl, yet it succeeded. (?)");
        }

        // Look for bad output, just to be sure (i.e., weird C output of \n\n)
        String tmp = "";
        if (this.poolindex.equals(tmp) ||
                this.localName.equals(tmp) ||
                this.primaryGroup.equals(tmp)) {
            throw new DynamicAccountsException("bad output from pool impl");
        }
        if (this.secondaryGroups != null) {
            for (int i = 0; i < this.secondaryGroups.length; i++) {
                if (this.secondaryGroups[i].equals(tmp)) {
                    throw new DynamicAccountsException("bad output from pool impl");
                }
            }
        }

        if (logger.isDebugEnabled()) {
            logger.debug("poolindex     :: " + this.poolindex);
            logger.debug("Local name    :: " + this.localName);
            logger.debug("Primary group :: " + this.primaryGroup);
            if (this.secondaryGroups != null) {
                for (int i = 0; i < this.secondaryGroups.length; i++) {
                    logger.debug("SecdGroup " + (i+1) + "  :: " + this.secondaryGroups[i]);
                }
            }
        }

    }

    protected void deleteAccountUsingMech() throws DynamicAccountsException {
        String command[] = null;

        if (this.poolindex == null) {
            throw new DynamicAccountsException("cannot delete, no poolindex");
        }

        command =  new String[3];
        command[0] = executable;
        command[1] = DELETE_BY_POOLINDEX_COMMAND;
        command[2] = this.poolindex;

        String returnString = null;
        try {
            returnString = DynamicAccountsUtil.runCommand(command);
        } catch (ReturnException e) {
            translateReturnValue(e);
        }

    }

    private void translateReturnValue(ReturnException e) throws DynamicAccountsException {
        /* #define GENERAL_FAILURE 255 */
        if (e.retval == 255) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("failure running pool" +
                    " account program\nSTDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("failure running pool" +
                        " account program");
            }
        }
        /* #define ENVIRONMENT 1 */
        if (e.retval == 1) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("pool account program" +
                    " setup problem\nSTDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("pool account program" +
                    " setup problem");
            }
        }
        /* #define SYNTAX 2 */
        if (e.retval == 2) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("pool account service" +
                    " setup problem\nSTDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("pool account service" +
                    "setup problem");
            }
        }
        /* #define BADARG 3 */
        if (e.retval == 3) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("problem with parameters, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("problem with parameters");
            }
        }
        /* #define POOL_INDEX_FAILURE 4 */
        if (e.retval == 4) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("problem with LCMAPS, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("problem with LCMAPS");
            }
        }
        /* #define NO_POOL_FOR_USER 5 */
        /* //not supported by LCMAPS
        if (e.retval == 5) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("found user for DN, but no pool account, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("found user for DN, but no pool account,");
            }
        } */
        /* #define NO_USER 6 */
        /* //not supported by LCMAPS
        if (e.retval == 6) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("no mapping for this DN, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("no mapping for this DN");
            }
        } */
        /* #define POOL_FULL 7 */
        /* //not supported by LCMAPS
        if (e.retval == 7) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("pool is exhausted, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("pool is exhausted");
            }
        } */



        /* This is returned if LCMAPS successfully initialized but the actual call to
           lcmaps_return_poolindex_without_gsi() failed.  This is in all likelihood
           because the dn+attr combination was denied a lease, *especially* if you have
           had at least one successful lease by LCMAPS
          */
        /* #define GRIDMAPDIR_NO_ASSOCIATION 8 */
        if (e.retval == 8) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("no lease available, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("no lease available");
            }
        }

        /*****
         * EXPIRY_PROG_FAILURE & EXPIRY_PROG_CHECK are both errors that happen
         * after a lease has been unleased and quarantined.  Perhaps don't throw
         * exception because this resource, which represents the lease, must succeed
         * in its destruction?
         */

        /* #define EXPIRY_PROG_FAILURE 9 */
        if (e.retval == 9) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("expire account program failure, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("expire account program failure");
            }
        }

        /* #define EXPIRY_PROG_CHECK 10 */
        if (e.retval == 10) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("expire account environment problem, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("expire account environment problem");
            }
        }

        /* #define NO_GROUP 11 */
        /* //not supported by LCMAPS
        if (e.retval == 11) {
            if (e.stderr != null) {
            throw new DynamicAccountsException("no group mapping for attribute, STDERR: " + e.stderr);
            } else {
            throw new DynamicAccountsException("no group mapping for attribute");
            }
        } */

        /* others? */
        if (e.stderr != null) {
        throw new DynamicAccountsException("pooling problem, implementation " +
                "returns error #" + e.retval + ", STDERR: " + e.stderr);
        } else {
        throw new DynamicAccountsException("pooling problem, implementation" +
                "returns error #" + e.retval);
        }

    }

    /*
     * END setuid interface related methods
     */


}
